/* Autogenerated file, DO NOT EDIT manually! generated by gen_perf.py
 *
 * Copyright (c) 2015 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#pragma once

struct intel_perf_config;

void intel_oa_register_queries_hsw(struct intel_perf_config *perf);
void intel_oa_register_queries_bdw(struct intel_perf_config *perf);
void intel_oa_register_queries_chv(struct intel_perf_config *perf);
void intel_oa_register_queries_sklgt2(struct intel_perf_config *perf);
void intel_oa_register_queries_sklgt3(struct intel_perf_config *perf);
void intel_oa_register_queries_sklgt4(struct intel_perf_config *perf);
void intel_oa_register_queries_kblgt2(struct intel_perf_config *perf);
void intel_oa_register_queries_kblgt3(struct intel_perf_config *perf);
void intel_oa_register_queries_cflgt2(struct intel_perf_config *perf);
void intel_oa_register_queries_cflgt3(struct intel_perf_config *perf);
void intel_oa_register_queries_bxt(struct intel_perf_config *perf);
void intel_oa_register_queries_glk(struct intel_perf_config *perf);
void intel_oa_register_queries_icl(struct intel_perf_config *perf);
void intel_oa_register_queries_ehl(struct intel_perf_config *perf);
void intel_oa_register_queries_tglgt1(struct intel_perf_config *perf);
void intel_oa_register_queries_tglgt2(struct intel_perf_config *perf);
void intel_oa_register_queries_rkl(struct intel_perf_config *perf);
void intel_oa_register_queries_dg1(struct intel_perf_config *perf);
void intel_oa_register_queries_adl(struct intel_perf_config *perf);
void intel_oa_register_queries_acmgt1(struct intel_perf_config *perf);
void intel_oa_register_queries_acmgt2(struct intel_perf_config *perf);
void intel_oa_register_queries_acmgt3(struct intel_perf_config *perf);
void intel_oa_register_queries_mtlgt2(struct intel_perf_config *perf);
void intel_oa_register_queries_mtlgt3(struct intel_perf_config *perf);
void intel_oa_register_queries_arlgt1(struct intel_perf_config *perf);
void intel_oa_register_queries_arlgt2(struct intel_perf_config *perf);
void intel_oa_register_queries_lnl(struct intel_perf_config *perf);
void intel_oa_register_queries_bmg(struct intel_perf_config *perf);
void intel_oa_register_queries_ptl(struct intel_perf_config *perf);
