#include "nv_push_clca6f.h"

#include <stdio.h>

#include "util/u_math.h"





const char *P_PARSE_NVC96F_MTHD(uint16_t idx) ATTRIBUTE_COLD;
void P_DUMP_NVC96F_MTHD_DATA(FILE *fp, uint16_t idx, uint32_t data,
                                   const char *prefix) ATTRIBUTE_COLD;

const char*
P_PARSE_NVCA6F_MTHD(uint16_t idx)
{
    switch (idx) {

  
        
    case NVCA6F_SEM_ADDR_LO:

        
    case NVCA6F_SEM_ADDR_HI:

        
    case NVCA6F_SEM_EXECUTE:


    return P_PARSE_NVC96F_MTHD(idx);

            
    default:
        return "unknown method";
    }
}

void
P_DUMP_NVCA6F_MTHD_DATA(FILE *fp, uint16_t idx, uint32_t data,
                         const char *prefix)
{
    UNUSED uint32_t parsed;
    switch (idx) {

  
        
    case NVCA6F_SEM_ADDR_LO:

        
    case NVCA6F_SEM_ADDR_HI:

        
    case NVCA6F_SEM_EXECUTE:


    P_DUMP_NVC96F_MTHD_DATA(fp, idx, data, prefix);
    break;

                default:
        fprintf(fp, "%s.VALUE = 0x%x\n", prefix, data);
        break;
    }
}
