#include "nv_push_cla06f.h"

#include <stdio.h>

#include "util/u_math.h"





const char *P_PARSE_NV906F_MTHD(uint16_t idx) ATTRIBUTE_COLD;
void P_DUMP_NV906F_MTHD_DATA(FILE *fp, uint16_t idx, uint32_t data,
                                   const char *prefix) ATTRIBUTE_COLD;

const char*
P_PARSE_NVA06F_MTHD(uint16_t idx)
{
    switch (idx) {

  
    default:

    return P_PARSE_NV906F_MTHD(idx);

                                                        
    }
}

void
P_DUMP_NVA06F_MTHD_DATA(FILE *fp, uint16_t idx, uint32_t data,
                         const char *prefix)
{
    UNUSED uint32_t parsed;
    switch (idx) {

  
    default:

    P_DUMP_NV906F_MTHD_DATA(fp, idx, data, prefix);
    break;

                                                            }
}
