
/* Copyright © 2015-2021 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/* This file generated from vk_entrypoints_gen.py, don't edit directly. */

/* This file is the only place we rely on undefined symbols to fall back to
 * NULL. Other files use regular symbol declarations.
 * See also comments on VK_ENTRY_WEAK.
 */
#define VK_ENTRY_USE_WEAK 1
#include "v3dv_entrypoints.h"

/* Weak aliases for all potential implementations. These will resolve to
 * NULL if they're not defined, which lets the resolve_entrypoint() function
 * either pick the correct entry point.
 *
 * MSVC uses different decorated names for 32-bit versus 64-bit. Declare
 * all argument sizes for 32-bit because computing the actual size would be
 * difficult.
 */




#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateInstance@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateInstance=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateInstance=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyInstance@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyInstance=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyInstance=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDevices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumeratePhysicalDevices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumeratePhysicalDevices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetInstanceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetInstanceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetInstanceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceVersion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumerateInstanceVersion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumerateInstanceVersion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceLayerProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumerateInstanceLayerProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumerateInstanceLayerProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateInstanceExtensionProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumerateInstanceExtensionProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumerateInstanceExtensionProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAndroidSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateAndroidSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateAndroidSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSurfaceOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateSurfaceOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateSurfaceOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayPlaneSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDisplayPlaneSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDisplayPlaneSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroySurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroySurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_VI_NN
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateViSurfaceNN@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateViSurfaceNN=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateViSurfaceNN=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_VI_NN
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWaylandSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateWaylandSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateWaylandSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateUbmSurfaceSEC@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateUbmSurfaceSEC=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateUbmSurfaceSEC=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateWin32SurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateWin32SurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateWin32SurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXlibSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateXlibSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateXlibSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateXcbSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateXcbSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateXcbSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDirectFBSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDirectFBSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDirectFBSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImagePipeSurfaceFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateImagePipeSurfaceFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateImagePipeSurfaceFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_GGP
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateStreamDescriptorSurfaceGGP@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateStreamDescriptorSurfaceGGP=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateStreamDescriptorSurfaceGGP=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_GGP
#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateScreenSurfaceQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateScreenSurfaceQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateScreenSurfaceQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugReportCallbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDebugReportCallbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDebugReportCallbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugReportCallbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyDebugReportCallbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyDebugReportCallbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugReportMessageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DebugReportMessageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DebugReportMessageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroups@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumeratePhysicalDeviceGroups=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumeratePhysicalDeviceGroups=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceGroupsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumeratePhysicalDeviceGroupsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumeratePhysicalDeviceGroupsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_IOS_MVK
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIOSSurfaceMVK@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateIOSSurfaceMVK=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateIOSSurfaceMVK=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_IOS_MVK
#ifdef VK_USE_PLATFORM_MACOS_MVK
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMacOSSurfaceMVK@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateMacOSSurfaceMVK=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateMacOSSurfaceMVK=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_MACOS_MVK
#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMetalSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateMetalSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateMetalSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDebugUtilsMessengerEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDebugUtilsMessengerEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDebugUtilsMessengerEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDebugUtilsMessengerEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyDebugUtilsMessengerEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyDebugUtilsMessengerEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SubmitDebugUtilsMessageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SubmitDebugUtilsMessageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SubmitDebugUtilsMessageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateHeadlessSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateHeadlessSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateHeadlessSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetExternalComputeQueueDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetExternalComputeQueueDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetExternalComputeQueueDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_instance_entrypoint_table v3dv_instance_entrypoints = {
    .CreateInstance = v3dv_CreateInstance,
    .DestroyInstance = v3dv_DestroyInstance,
    .EnumeratePhysicalDevices = v3dv_EnumeratePhysicalDevices,
    .GetInstanceProcAddr = v3dv_GetInstanceProcAddr,
    .EnumerateInstanceVersion = v3dv_EnumerateInstanceVersion,
    .EnumerateInstanceLayerProperties = v3dv_EnumerateInstanceLayerProperties,
    .EnumerateInstanceExtensionProperties = v3dv_EnumerateInstanceExtensionProperties,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .CreateAndroidSurfaceKHR = v3dv_CreateAndroidSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateAndroidSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_OHOS
    .CreateSurfaceOHOS = v3dv_CreateSurfaceOHOS,
#elif defined(_MSC_VER)
    .CreateSurfaceOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .CreateDisplayPlaneSurfaceKHR = v3dv_CreateDisplayPlaneSurfaceKHR,
    .DestroySurfaceKHR = v3dv_DestroySurfaceKHR,
#ifdef VK_USE_PLATFORM_VI_NN
    .CreateViSurfaceNN = v3dv_CreateViSurfaceNN,
#elif defined(_MSC_VER)
    .CreateViSurfaceNN = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_VI_NN
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
    .CreateWaylandSurfaceKHR = v3dv_CreateWaylandSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateWaylandSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
    .CreateUbmSurfaceSEC = v3dv_CreateUbmSurfaceSEC,
#elif defined(_MSC_VER)
    .CreateUbmSurfaceSEC = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .CreateWin32SurfaceKHR = v3dv_CreateWin32SurfaceKHR,
#elif defined(_MSC_VER)
    .CreateWin32SurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
    .CreateXlibSurfaceKHR = v3dv_CreateXlibSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateXlibSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
    .CreateXcbSurfaceKHR = v3dv_CreateXcbSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateXcbSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
    .CreateDirectFBSurfaceEXT = v3dv_CreateDirectFBSurfaceEXT,
#elif defined(_MSC_VER)
    .CreateDirectFBSurfaceEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateImagePipeSurfaceFUCHSIA = v3dv_CreateImagePipeSurfaceFUCHSIA,
#elif defined(_MSC_VER)
    .CreateImagePipeSurfaceFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_GGP
    .CreateStreamDescriptorSurfaceGGP = v3dv_CreateStreamDescriptorSurfaceGGP,
#elif defined(_MSC_VER)
    .CreateStreamDescriptorSurfaceGGP = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_GGP
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .CreateScreenSurfaceQNX = v3dv_CreateScreenSurfaceQNX,
#elif defined(_MSC_VER)
    .CreateScreenSurfaceQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CreateDebugReportCallbackEXT = v3dv_CreateDebugReportCallbackEXT,
    .DestroyDebugReportCallbackEXT = v3dv_DestroyDebugReportCallbackEXT,
    .DebugReportMessageEXT = v3dv_DebugReportMessageEXT,
    .EnumeratePhysicalDeviceGroups = v3dv_EnumeratePhysicalDeviceGroups,
    .EnumeratePhysicalDeviceGroupsKHR = v3dv_EnumeratePhysicalDeviceGroupsKHR,
#ifdef VK_USE_PLATFORM_IOS_MVK
    .CreateIOSSurfaceMVK = v3dv_CreateIOSSurfaceMVK,
#elif defined(_MSC_VER)
    .CreateIOSSurfaceMVK = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_IOS_MVK
#ifdef VK_USE_PLATFORM_MACOS_MVK
    .CreateMacOSSurfaceMVK = v3dv_CreateMacOSSurfaceMVK,
#elif defined(_MSC_VER)
    .CreateMacOSSurfaceMVK = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_MACOS_MVK
#ifdef VK_USE_PLATFORM_METAL_EXT
    .CreateMetalSurfaceEXT = v3dv_CreateMetalSurfaceEXT,
#elif defined(_MSC_VER)
    .CreateMetalSurfaceEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CreateDebugUtilsMessengerEXT = v3dv_CreateDebugUtilsMessengerEXT,
    .DestroyDebugUtilsMessengerEXT = v3dv_DestroyDebugUtilsMessengerEXT,
    .SubmitDebugUtilsMessageEXT = v3dv_SubmitDebugUtilsMessageEXT,
    .CreateHeadlessSurfaceEXT = v3dv_CreateHeadlessSurfaceEXT,
    .GetExternalComputeQueueDataNV = v3dv_GetExternalComputeQueueDataNV,
};



#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceQueueFamilyProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceQueueFamilyProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceMemoryProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceMemoryProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceImageFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceImageFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceLayerProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumerateDeviceLayerProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumerateDeviceLayerProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumerateDeviceExtensionProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumerateDeviceExtensionProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumerateDeviceExtensionProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSparseImageFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSparseImageFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceDisplayPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceDisplayPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneSupportedDisplaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDisplayPlaneSupportedDisplaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDisplayPlaneSupportedDisplaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDisplayModePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDisplayModePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDisplayModeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDisplayModeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDisplayModeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDisplayPlaneCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDisplayPlaneCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSurfaceSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSurfaceSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormatsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSurfaceFormatsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSurfaceFormatsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WAYLAND_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceUbmPresentationSupportSEC@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceUbmPresentationSupportSEC=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceUbmPresentationSupportSEC=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceWin32PresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceWin32PresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceWin32PresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXlibPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceXlibPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceXlibPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceXcbPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceXcbPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceXcbPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceScreenPresentationSupportQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceScreenPresentationSupportQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceScreenPresentationSupportQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceFeatures2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceFeatures2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFeatures2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceFeatures2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceFeatures2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceImageFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceImageFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceImageFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceImageFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceImageFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceQueueFamilyProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceQueueFamilyProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceMemoryProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceMemoryProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMemoryProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceMemoryProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceMemoryProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSparseImageFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSparseImageFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceExternalBufferProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceExternalBufferProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphoreProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceExternalSemaphoreProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceExternalSemaphoreProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFenceProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceExternalFenceProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceExternalFenceProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalFencePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceExternalFencePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceExternalFencePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ReleaseDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ReleaseDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireXlibDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AcquireXlibDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AcquireXlibDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRandROutputDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetRandROutputDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetRandROutputDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireWinrtDisplayNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AcquireWinrtDisplayNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AcquireWinrtDisplayNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetWinrtDisplayNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetWinrtDisplayNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetWinrtDisplayNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDevicePresentRectanglesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDevicePresentRectanglesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDevicePresentRectanglesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceMultisamplePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceMultisamplePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceMultisamplePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfaceFormats2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSurfaceFormats2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSurfaceFormats2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceDisplayProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceDisplayProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayModeProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDisplayModeProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDisplayModeProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDisplayPlaneCapabilities2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDisplayPlaneCapabilities2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDisplayPlaneCapabilities2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceToolProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceToolProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceToolPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceToolPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceToolPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceFragmentShadingRatesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceFragmentShadingRatesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceFragmentShadingRatesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceVideoCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceVideoCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireDrmDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AcquireDrmDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AcquireDrmDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDrmDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDrmDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDrmDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceExternalTensorPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceExternalTensorPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceExternalTensorPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceDescriptorSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceDescriptorSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceDescriptorSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_physical_device_entrypoint_table v3dv_physical_device_entrypoints = {
    .GetPhysicalDeviceProperties = v3dv_GetPhysicalDeviceProperties,
    .GetPhysicalDeviceQueueFamilyProperties = v3dv_GetPhysicalDeviceQueueFamilyProperties,
    .GetPhysicalDeviceMemoryProperties = v3dv_GetPhysicalDeviceMemoryProperties,
    .GetPhysicalDeviceFeatures = v3dv_GetPhysicalDeviceFeatures,
    .GetPhysicalDeviceFormatProperties = v3dv_GetPhysicalDeviceFormatProperties,
    .GetPhysicalDeviceImageFormatProperties = v3dv_GetPhysicalDeviceImageFormatProperties,
    .CreateDevice = v3dv_CreateDevice,
    .EnumerateDeviceLayerProperties = v3dv_EnumerateDeviceLayerProperties,
    .EnumerateDeviceExtensionProperties = v3dv_EnumerateDeviceExtensionProperties,
    .GetPhysicalDeviceSparseImageFormatProperties = v3dv_GetPhysicalDeviceSparseImageFormatProperties,
    .GetPhysicalDeviceDisplayPropertiesKHR = v3dv_GetPhysicalDeviceDisplayPropertiesKHR,
    .GetPhysicalDeviceDisplayPlanePropertiesKHR = v3dv_GetPhysicalDeviceDisplayPlanePropertiesKHR,
    .GetDisplayPlaneSupportedDisplaysKHR = v3dv_GetDisplayPlaneSupportedDisplaysKHR,
    .GetDisplayModePropertiesKHR = v3dv_GetDisplayModePropertiesKHR,
    .CreateDisplayModeKHR = v3dv_CreateDisplayModeKHR,
    .GetDisplayPlaneCapabilitiesKHR = v3dv_GetDisplayPlaneCapabilitiesKHR,
    .GetPhysicalDeviceSurfaceSupportKHR = v3dv_GetPhysicalDeviceSurfaceSupportKHR,
    .GetPhysicalDeviceSurfaceCapabilitiesKHR = v3dv_GetPhysicalDeviceSurfaceCapabilitiesKHR,
    .GetPhysicalDeviceSurfaceFormatsKHR = v3dv_GetPhysicalDeviceSurfaceFormatsKHR,
    .GetPhysicalDeviceSurfacePresentModesKHR = v3dv_GetPhysicalDeviceSurfacePresentModesKHR,
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
    .GetPhysicalDeviceWaylandPresentationSupportKHR = v3dv_GetPhysicalDeviceWaylandPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceWaylandPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
    .GetPhysicalDeviceUbmPresentationSupportSEC = v3dv_GetPhysicalDeviceUbmPresentationSupportSEC,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceUbmPresentationSupportSEC = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceWin32PresentationSupportKHR = v3dv_GetPhysicalDeviceWin32PresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceWin32PresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
    .GetPhysicalDeviceXlibPresentationSupportKHR = v3dv_GetPhysicalDeviceXlibPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceXlibPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
    .GetPhysicalDeviceXcbPresentationSupportKHR = v3dv_GetPhysicalDeviceXcbPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceXcbPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
    .GetPhysicalDeviceDirectFBPresentationSupportEXT = v3dv_GetPhysicalDeviceDirectFBPresentationSupportEXT,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceDirectFBPresentationSupportEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetPhysicalDeviceScreenPresentationSupportQNX = v3dv_GetPhysicalDeviceScreenPresentationSupportQNX,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceScreenPresentationSupportQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .GetPhysicalDeviceExternalImageFormatPropertiesNV = v3dv_GetPhysicalDeviceExternalImageFormatPropertiesNV,
    .GetPhysicalDeviceFeatures2 = v3dv_GetPhysicalDeviceFeatures2,
    .GetPhysicalDeviceFeatures2KHR = v3dv_GetPhysicalDeviceFeatures2KHR,
    .GetPhysicalDeviceProperties2 = v3dv_GetPhysicalDeviceProperties2,
    .GetPhysicalDeviceProperties2KHR = v3dv_GetPhysicalDeviceProperties2KHR,
    .GetPhysicalDeviceFormatProperties2 = v3dv_GetPhysicalDeviceFormatProperties2,
    .GetPhysicalDeviceFormatProperties2KHR = v3dv_GetPhysicalDeviceFormatProperties2KHR,
    .GetPhysicalDeviceImageFormatProperties2 = v3dv_GetPhysicalDeviceImageFormatProperties2,
    .GetPhysicalDeviceImageFormatProperties2KHR = v3dv_GetPhysicalDeviceImageFormatProperties2KHR,
    .GetPhysicalDeviceQueueFamilyProperties2 = v3dv_GetPhysicalDeviceQueueFamilyProperties2,
    .GetPhysicalDeviceQueueFamilyProperties2KHR = v3dv_GetPhysicalDeviceQueueFamilyProperties2KHR,
    .GetPhysicalDeviceMemoryProperties2 = v3dv_GetPhysicalDeviceMemoryProperties2,
    .GetPhysicalDeviceMemoryProperties2KHR = v3dv_GetPhysicalDeviceMemoryProperties2KHR,
    .GetPhysicalDeviceSparseImageFormatProperties2 = v3dv_GetPhysicalDeviceSparseImageFormatProperties2,
    .GetPhysicalDeviceSparseImageFormatProperties2KHR = v3dv_GetPhysicalDeviceSparseImageFormatProperties2KHR,
    .GetPhysicalDeviceExternalBufferProperties = v3dv_GetPhysicalDeviceExternalBufferProperties,
    .GetPhysicalDeviceExternalBufferPropertiesKHR = v3dv_GetPhysicalDeviceExternalBufferPropertiesKHR,
    .GetPhysicalDeviceExternalSemaphoreProperties = v3dv_GetPhysicalDeviceExternalSemaphoreProperties,
    .GetPhysicalDeviceExternalSemaphorePropertiesKHR = v3dv_GetPhysicalDeviceExternalSemaphorePropertiesKHR,
    .GetPhysicalDeviceExternalFenceProperties = v3dv_GetPhysicalDeviceExternalFenceProperties,
    .GetPhysicalDeviceExternalFencePropertiesKHR = v3dv_GetPhysicalDeviceExternalFencePropertiesKHR,
    .ReleaseDisplayEXT = v3dv_ReleaseDisplayEXT,
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
    .AcquireXlibDisplayEXT = v3dv_AcquireXlibDisplayEXT,
#elif defined(_MSC_VER)
    .AcquireXlibDisplayEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
    .GetRandROutputDisplayEXT = v3dv_GetRandROutputDisplayEXT,
#elif defined(_MSC_VER)
    .GetRandROutputDisplayEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireWinrtDisplayNV = v3dv_AcquireWinrtDisplayNV,
#elif defined(_MSC_VER)
    .AcquireWinrtDisplayNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetWinrtDisplayNV = v3dv_GetWinrtDisplayNV,
#elif defined(_MSC_VER)
    .GetWinrtDisplayNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceSurfaceCapabilities2EXT = v3dv_GetPhysicalDeviceSurfaceCapabilities2EXT,
    .GetPhysicalDevicePresentRectanglesKHR = v3dv_GetPhysicalDevicePresentRectanglesKHR,
    .GetPhysicalDeviceMultisamplePropertiesEXT = v3dv_GetPhysicalDeviceMultisamplePropertiesEXT,
    .GetPhysicalDeviceSurfaceCapabilities2KHR = v3dv_GetPhysicalDeviceSurfaceCapabilities2KHR,
    .GetPhysicalDeviceSurfaceFormats2KHR = v3dv_GetPhysicalDeviceSurfaceFormats2KHR,
    .GetPhysicalDeviceDisplayProperties2KHR = v3dv_GetPhysicalDeviceDisplayProperties2KHR,
    .GetPhysicalDeviceDisplayPlaneProperties2KHR = v3dv_GetPhysicalDeviceDisplayPlaneProperties2KHR,
    .GetDisplayModeProperties2KHR = v3dv_GetDisplayModeProperties2KHR,
    .GetDisplayPlaneCapabilities2KHR = v3dv_GetDisplayPlaneCapabilities2KHR,
    .GetPhysicalDeviceCalibrateableTimeDomainsKHR = v3dv_GetPhysicalDeviceCalibrateableTimeDomainsKHR,
    .GetPhysicalDeviceCalibrateableTimeDomainsEXT = v3dv_GetPhysicalDeviceCalibrateableTimeDomainsEXT,
    .GetPhysicalDeviceCooperativeMatrixPropertiesNV = v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceSurfacePresentModes2EXT = v3dv_GetPhysicalDeviceSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR = v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR,
    .GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR = v3dv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR,
    .GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV = v3dv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV,
    .GetPhysicalDeviceToolProperties = v3dv_GetPhysicalDeviceToolProperties,
    .GetPhysicalDeviceToolPropertiesEXT = v3dv_GetPhysicalDeviceToolPropertiesEXT,
    .GetPhysicalDeviceFragmentShadingRatesKHR = v3dv_GetPhysicalDeviceFragmentShadingRatesKHR,
    .GetPhysicalDeviceVideoCapabilitiesKHR = v3dv_GetPhysicalDeviceVideoCapabilitiesKHR,
    .GetPhysicalDeviceVideoFormatPropertiesKHR = v3dv_GetPhysicalDeviceVideoFormatPropertiesKHR,
    .GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR = v3dv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR,
    .AcquireDrmDisplayEXT = v3dv_AcquireDrmDisplayEXT,
    .GetDrmDisplayEXT = v3dv_GetDrmDisplayEXT,
    .GetPhysicalDeviceOpticalFlowImageFormatsNV = v3dv_GetPhysicalDeviceOpticalFlowImageFormatsNV,
    .GetPhysicalDeviceCooperativeMatrixPropertiesKHR = v3dv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR,
    .GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV = v3dv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV,
    .GetPhysicalDeviceCooperativeVectorPropertiesNV = v3dv_GetPhysicalDeviceCooperativeVectorPropertiesNV,
    .EnumeratePhysicalDeviceShaderInstrumentationMetricsARM = v3dv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM,
    .GetPhysicalDeviceExternalTensorPropertiesARM = v3dv_GetPhysicalDeviceExternalTensorPropertiesARM,
    .GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM = v3dv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM,
    .GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM = v3dv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM,
    .EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM = v3dv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM,
    .GetPhysicalDeviceDescriptorSizeEXT = v3dv_GetPhysicalDeviceDescriptorSizeEXT,
    .GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM = v3dv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM,
};



#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_v3dv_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:v3dv_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#v3dv_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver42_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver42_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver42_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_ver71_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:ver71_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#ver71_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_device_entrypoint_table v3dv_device_entrypoints = {
    .GetDeviceProcAddr = v3dv_GetDeviceProcAddr,
    .DestroyDevice = v3dv_DestroyDevice,
    .GetDeviceQueue = v3dv_GetDeviceQueue,
    .QueueSubmit = v3dv_QueueSubmit,
    .QueueWaitIdle = v3dv_QueueWaitIdle,
    .DeviceWaitIdle = v3dv_DeviceWaitIdle,
    .AllocateMemory = v3dv_AllocateMemory,
    .FreeMemory = v3dv_FreeMemory,
    .MapMemory = v3dv_MapMemory,
    .UnmapMemory = v3dv_UnmapMemory,
    .FlushMappedMemoryRanges = v3dv_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = v3dv_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = v3dv_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = v3dv_GetBufferMemoryRequirements,
    .BindBufferMemory = v3dv_BindBufferMemory,
    .GetImageMemoryRequirements = v3dv_GetImageMemoryRequirements,
    .BindImageMemory = v3dv_BindImageMemory,
    .GetImageSparseMemoryRequirements = v3dv_GetImageSparseMemoryRequirements,
    .QueueBindSparse = v3dv_QueueBindSparse,
    .CreateFence = v3dv_CreateFence,
    .DestroyFence = v3dv_DestroyFence,
    .ResetFences = v3dv_ResetFences,
    .GetFenceStatus = v3dv_GetFenceStatus,
    .WaitForFences = v3dv_WaitForFences,
    .CreateSemaphore = v3dv_CreateSemaphore,
    .DestroySemaphore = v3dv_DestroySemaphore,
    .CreateEvent = v3dv_CreateEvent,
    .DestroyEvent = v3dv_DestroyEvent,
    .GetEventStatus = v3dv_GetEventStatus,
    .SetEvent = v3dv_SetEvent,
    .ResetEvent = v3dv_ResetEvent,
    .CreateQueryPool = v3dv_CreateQueryPool,
    .DestroyQueryPool = v3dv_DestroyQueryPool,
    .GetQueryPoolResults = v3dv_GetQueryPoolResults,
    .ResetQueryPool = v3dv_ResetQueryPool,
    .ResetQueryPoolEXT = v3dv_ResetQueryPoolEXT,
    .CreateBuffer = v3dv_CreateBuffer,
    .DestroyBuffer = v3dv_DestroyBuffer,
    .CreateBufferView = v3dv_CreateBufferView,
    .DestroyBufferView = v3dv_DestroyBufferView,
    .CreateImage = v3dv_CreateImage,
    .DestroyImage = v3dv_DestroyImage,
    .GetImageSubresourceLayout = v3dv_GetImageSubresourceLayout,
    .CreateImageView = v3dv_CreateImageView,
    .DestroyImageView = v3dv_DestroyImageView,
    .CreateShaderModule = v3dv_CreateShaderModule,
    .DestroyShaderModule = v3dv_DestroyShaderModule,
    .CreatePipelineCache = v3dv_CreatePipelineCache,
    .DestroyPipelineCache = v3dv_DestroyPipelineCache,
    .GetPipelineCacheData = v3dv_GetPipelineCacheData,
    .MergePipelineCaches = v3dv_MergePipelineCaches,
    .CreatePipelineBinariesKHR = v3dv_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = v3dv_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = v3dv_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = v3dv_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = v3dv_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = v3dv_CreateGraphicsPipelines,
    .CreateComputePipelines = v3dv_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = v3dv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = v3dv_DestroyPipeline,
    .CreatePipelineLayout = v3dv_CreatePipelineLayout,
    .DestroyPipelineLayout = v3dv_DestroyPipelineLayout,
    .CreateSampler = v3dv_CreateSampler,
    .DestroySampler = v3dv_DestroySampler,
    .CreateDescriptorSetLayout = v3dv_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = v3dv_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = v3dv_CreateDescriptorPool,
    .DestroyDescriptorPool = v3dv_DestroyDescriptorPool,
    .ResetDescriptorPool = v3dv_ResetDescriptorPool,
    .AllocateDescriptorSets = v3dv_AllocateDescriptorSets,
    .FreeDescriptorSets = v3dv_FreeDescriptorSets,
    .UpdateDescriptorSets = v3dv_UpdateDescriptorSets,
    .CreateFramebuffer = v3dv_CreateFramebuffer,
    .DestroyFramebuffer = v3dv_DestroyFramebuffer,
    .CreateRenderPass = v3dv_CreateRenderPass,
    .DestroyRenderPass = v3dv_DestroyRenderPass,
    .GetRenderAreaGranularity = v3dv_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = v3dv_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = v3dv_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = v3dv_CreateCommandPool,
    .DestroyCommandPool = v3dv_DestroyCommandPool,
    .ResetCommandPool = v3dv_ResetCommandPool,
    .AllocateCommandBuffers = v3dv_AllocateCommandBuffers,
    .FreeCommandBuffers = v3dv_FreeCommandBuffers,
    .BeginCommandBuffer = v3dv_BeginCommandBuffer,
    .EndCommandBuffer = v3dv_EndCommandBuffer,
    .ResetCommandBuffer = v3dv_ResetCommandBuffer,
    .CmdBindPipeline = v3dv_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = v3dv_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = v3dv_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = v3dv_CmdSetViewport,
    .CmdSetScissor = v3dv_CmdSetScissor,
    .CmdSetLineWidth = v3dv_CmdSetLineWidth,
    .CmdSetDepthBias = v3dv_CmdSetDepthBias,
    .CmdSetBlendConstants = v3dv_CmdSetBlendConstants,
    .CmdSetDepthBounds = v3dv_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = v3dv_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = v3dv_CmdSetStencilWriteMask,
    .CmdSetStencilReference = v3dv_CmdSetStencilReference,
    .CmdBindDescriptorSets = v3dv_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = v3dv_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = v3dv_CmdBindVertexBuffers,
    .CmdDraw = v3dv_CmdDraw,
    .CmdDrawIndexed = v3dv_CmdDrawIndexed,
    .CmdDrawMultiEXT = v3dv_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = v3dv_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = v3dv_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = v3dv_CmdDrawIndexedIndirect,
    .CmdDispatch = v3dv_CmdDispatch,
    .CmdDispatchIndirect = v3dv_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = v3dv_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = v3dv_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = v3dv_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = v3dv_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = v3dv_CmdCopyBuffer,
    .CmdCopyImage = v3dv_CmdCopyImage,
    .CmdBlitImage = v3dv_CmdBlitImage,
    .CmdCopyBufferToImage = v3dv_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = v3dv_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = v3dv_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = v3dv_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = v3dv_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = v3dv_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = v3dv_CmdUpdateBuffer,
    .CmdFillBuffer = v3dv_CmdFillBuffer,
    .CmdClearColorImage = v3dv_CmdClearColorImage,
    .CmdClearDepthStencilImage = v3dv_CmdClearDepthStencilImage,
    .CmdClearAttachments = v3dv_CmdClearAttachments,
    .CmdResolveImage = v3dv_CmdResolveImage,
    .CmdSetEvent = v3dv_CmdSetEvent,
    .CmdResetEvent = v3dv_CmdResetEvent,
    .CmdWaitEvents = v3dv_CmdWaitEvents,
    .CmdPipelineBarrier = v3dv_CmdPipelineBarrier,
    .CmdBeginQuery = v3dv_CmdBeginQuery,
    .CmdEndQuery = v3dv_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = v3dv_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = v3dv_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = v3dv_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = v3dv_CmdResetQueryPool,
    .CmdWriteTimestamp = v3dv_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = v3dv_CmdCopyQueryPoolResults,
    .CmdPushConstants = v3dv_CmdPushConstants,
    .CmdBeginRenderPass = v3dv_CmdBeginRenderPass,
    .CmdNextSubpass = v3dv_CmdNextSubpass,
    .CmdEndRenderPass = v3dv_CmdEndRenderPass,
    .CmdExecuteCommands = v3dv_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = v3dv_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = v3dv_CreateSwapchainKHR,
    .DestroySwapchainKHR = v3dv_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = v3dv_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = v3dv_AcquireNextImageKHR,
    .QueuePresentKHR = v3dv_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = v3dv_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = v3dv_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = v3dv_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = v3dv_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = v3dv_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = v3dv_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = v3dv_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = v3dv_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = v3dv_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = v3dv_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = v3dv_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = v3dv_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = v3dv_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = v3dv_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = v3dv_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = v3dv_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = v3dv_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = v3dv_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = v3dv_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = v3dv_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = v3dv_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = v3dv_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = v3dv_CmdPushDescriptorSetKHR,
    .TrimCommandPool = v3dv_TrimCommandPool,
    .TrimCommandPoolKHR = v3dv_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = v3dv_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = v3dv_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = v3dv_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = v3dv_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = v3dv_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = v3dv_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = v3dv_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = v3dv_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = v3dv_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = v3dv_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = v3dv_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = v3dv_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = v3dv_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = v3dv_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = v3dv_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = v3dv_GetFenceFdKHR,
    .ImportFenceFdKHR = v3dv_ImportFenceFdKHR,
    .DisplayPowerControlEXT = v3dv_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = v3dv_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = v3dv_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = v3dv_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = v3dv_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = v3dv_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = v3dv_BindBufferMemory2,
    .BindBufferMemory2KHR = v3dv_BindBufferMemory2KHR,
    .BindImageMemory2 = v3dv_BindImageMemory2,
    .BindImageMemory2KHR = v3dv_BindImageMemory2KHR,
    .CmdSetDeviceMask = v3dv_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = v3dv_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = v3dv_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = v3dv_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = v3dv_AcquireNextImage2KHR,
    .CmdDispatchBase = v3dv_CmdDispatchBase,
    .CmdDispatchBaseKHR = v3dv_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = v3dv_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = v3dv_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = v3dv_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = v3dv_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = v3dv_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = v3dv_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = v3dv_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = v3dv_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = v3dv_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = v3dv_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = v3dv_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = v3dv_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = v3dv_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = v3dv_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = v3dv_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = v3dv_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = v3dv_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = v3dv_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = v3dv_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = v3dv_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = v3dv_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = v3dv_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = v3dv_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = v3dv_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = v3dv_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = v3dv_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = v3dv_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = v3dv_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = v3dv_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = v3dv_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = v3dv_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = v3dv_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = v3dv_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = v3dv_GetDeviceQueue2,
    .CreateValidationCacheEXT = v3dv_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = v3dv_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = v3dv_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = v3dv_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = v3dv_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = v3dv_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = v3dv_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = v3dv_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = v3dv_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = v3dv_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = v3dv_GetShaderInfoAMD,
    .SetLocalDimmingAMD = v3dv_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = v3dv_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = v3dv_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = v3dv_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = v3dv_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = v3dv_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = v3dv_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = v3dv_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = v3dv_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = v3dv_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = v3dv_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = v3dv_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = v3dv_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = v3dv_CreateRenderPass2,
    .CreateRenderPass2KHR = v3dv_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = v3dv_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = v3dv_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = v3dv_CmdNextSubpass2,
    .CmdNextSubpass2KHR = v3dv_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = v3dv_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = v3dv_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = v3dv_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = v3dv_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = v3dv_WaitSemaphores,
    .WaitSemaphoresKHR = v3dv_WaitSemaphoresKHR,
    .SignalSemaphore = v3dv_SignalSemaphore,
    .SignalSemaphoreKHR = v3dv_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = v3dv_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = v3dv_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = v3dv_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = v3dv_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = v3dv_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = v3dv_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = v3dv_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = v3dv_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = v3dv_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = v3dv_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = v3dv_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = v3dv_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = v3dv_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = v3dv_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = v3dv_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = v3dv_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = v3dv_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = v3dv_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = v3dv_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = v3dv_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = v3dv_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = v3dv_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = v3dv_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = v3dv_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = v3dv_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = v3dv_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = v3dv_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = v3dv_CompileDeferredNV,
    .CreateAccelerationStructureNV = v3dv_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = v3dv_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = v3dv_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = v3dv_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = v3dv_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = v3dv_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = v3dv_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = v3dv_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = v3dv_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = v3dv_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = v3dv_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = v3dv_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = v3dv_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = v3dv_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = v3dv_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = v3dv_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = v3dv_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = v3dv_CmdTraceRaysKHR,
    .CmdTraceRaysNV = v3dv_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = v3dv_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = v3dv_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = v3dv_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = v3dv_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = v3dv_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = v3dv_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = v3dv_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = v3dv_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = v3dv_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = v3dv_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = v3dv_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = v3dv_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = v3dv_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = v3dv_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = v3dv_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = v3dv_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = v3dv_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = v3dv_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = v3dv_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = v3dv_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = v3dv_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = v3dv_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = v3dv_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = v3dv_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = v3dv_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = v3dv_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = v3dv_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = v3dv_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = v3dv_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = v3dv_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = v3dv_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = v3dv_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = v3dv_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = v3dv_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = v3dv_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = v3dv_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = v3dv_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = v3dv_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = v3dv_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = v3dv_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = v3dv_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = v3dv_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = v3dv_CmdSetLineStipple,
    .CmdSetLineStippleKHR = v3dv_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = v3dv_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = v3dv_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = v3dv_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = v3dv_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = v3dv_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = v3dv_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = v3dv_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = v3dv_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = v3dv_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = v3dv_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = v3dv_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = v3dv_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = v3dv_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = v3dv_AntiLagUpdateAMD,
    .CmdSetCullMode = v3dv_CmdSetCullMode,
    .CmdSetCullModeEXT = v3dv_CmdSetCullModeEXT,
    .CmdSetFrontFace = v3dv_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = v3dv_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = v3dv_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = v3dv_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = v3dv_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = v3dv_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = v3dv_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = v3dv_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = v3dv_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = v3dv_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = v3dv_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = v3dv_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = v3dv_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = v3dv_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = v3dv_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = v3dv_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = v3dv_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = v3dv_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = v3dv_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = v3dv_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = v3dv_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = v3dv_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = v3dv_CmdSetStencilOp,
    .CmdSetStencilOpEXT = v3dv_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = v3dv_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = v3dv_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = v3dv_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = v3dv_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = v3dv_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = v3dv_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = v3dv_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = v3dv_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = v3dv_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = v3dv_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = v3dv_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = v3dv_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = v3dv_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = v3dv_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = v3dv_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = v3dv_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = v3dv_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = v3dv_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = v3dv_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = v3dv_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = v3dv_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = v3dv_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = v3dv_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = v3dv_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = v3dv_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = v3dv_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = v3dv_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = v3dv_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = v3dv_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = v3dv_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = v3dv_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = v3dv_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = v3dv_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = v3dv_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = v3dv_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = v3dv_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = v3dv_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = v3dv_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = v3dv_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = v3dv_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = v3dv_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = v3dv_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = v3dv_DestroyPrivateDataSlotEXT,
    .SetPrivateData = v3dv_SetPrivateData,
    .SetPrivateDataEXT = v3dv_SetPrivateDataEXT,
    .GetPrivateData = v3dv_GetPrivateData,
    .GetPrivateDataEXT = v3dv_GetPrivateDataEXT,
    .CmdCopyBuffer2 = v3dv_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = v3dv_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = v3dv_CmdCopyImage2,
    .CmdCopyImage2KHR = v3dv_CmdCopyImage2KHR,
    .CmdBlitImage2 = v3dv_CmdBlitImage2,
    .CmdBlitImage2KHR = v3dv_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = v3dv_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = v3dv_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = v3dv_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = v3dv_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = v3dv_CmdResolveImage2,
    .CmdResolveImage2KHR = v3dv_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = v3dv_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = v3dv_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = v3dv_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = v3dv_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = v3dv_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = v3dv_CmdSetEvent2,
    .CmdSetEvent2KHR = v3dv_CmdSetEvent2KHR,
    .CmdResetEvent2 = v3dv_CmdResetEvent2,
    .CmdResetEvent2KHR = v3dv_CmdResetEvent2KHR,
    .CmdWaitEvents2 = v3dv_CmdWaitEvents2,
    .CmdWaitEvents2KHR = v3dv_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = v3dv_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = v3dv_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = v3dv_QueueSubmit2,
    .QueueSubmit2KHR = v3dv_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = v3dv_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = v3dv_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = v3dv_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = v3dv_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = v3dv_CopyMemoryToImage,
    .CopyMemoryToImageEXT = v3dv_CopyMemoryToImageEXT,
    .CopyImageToMemory = v3dv_CopyImageToMemory,
    .CopyImageToMemoryEXT = v3dv_CopyImageToMemoryEXT,
    .CopyImageToImage = v3dv_CopyImageToImage,
    .CopyImageToImageEXT = v3dv_CopyImageToImageEXT,
    .TransitionImageLayout = v3dv_TransitionImageLayout,
    .TransitionImageLayoutEXT = v3dv_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = v3dv_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = v3dv_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = v3dv_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = v3dv_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = v3dv_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = v3dv_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = v3dv_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = v3dv_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = v3dv_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = v3dv_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = v3dv_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = v3dv_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = v3dv_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = v3dv_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = v3dv_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = v3dv_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = v3dv_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = v3dv_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = v3dv_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = v3dv_CreateCuModuleNVX,
    .CreateCuFunctionNVX = v3dv_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = v3dv_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = v3dv_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = v3dv_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = v3dv_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = v3dv_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = v3dv_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = v3dv_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = v3dv_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = v3dv_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = v3dv_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = v3dv_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = v3dv_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = v3dv_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = v3dv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = v3dv_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = v3dv_WaitForPresent2KHR,
    .WaitForPresentKHR = v3dv_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = v3dv_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = v3dv_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = v3dv_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = v3dv_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = v3dv_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = v3dv_CmdBeginRendering,
    .CmdBeginRenderingKHR = v3dv_CmdBeginRenderingKHR,
    .CmdEndRendering = v3dv_CmdEndRendering,
    .CmdEndRendering2KHR = v3dv_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = v3dv_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = v3dv_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = v3dv_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = v3dv_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = v3dv_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = v3dv_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = v3dv_BuildMicromapsEXT,
    .DestroyMicromapEXT = v3dv_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = v3dv_CmdCopyMicromapEXT,
    .CopyMicromapEXT = v3dv_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = v3dv_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = v3dv_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = v3dv_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = v3dv_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = v3dv_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = v3dv_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = v3dv_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = v3dv_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = v3dv_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = v3dv_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = v3dv_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = v3dv_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = v3dv_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = v3dv_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = v3dv_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = v3dv_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = v3dv_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = v3dv_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = v3dv_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = v3dv_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = v3dv_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = v3dv_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = v3dv_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = v3dv_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = v3dv_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = v3dv_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = v3dv_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = v3dv_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = v3dv_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = v3dv_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = v3dv_MapMemory2,
    .MapMemory2KHR = v3dv_MapMemory2KHR,
    .UnmapMemory2 = v3dv_UnmapMemory2,
    .UnmapMemory2KHR = v3dv_UnmapMemory2KHR,
    .CreateShadersEXT = v3dv_CreateShadersEXT,
    .DestroyShaderEXT = v3dv_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = v3dv_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = v3dv_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = v3dv_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = v3dv_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = v3dv_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = v3dv_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = v3dv_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = v3dv_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = v3dv_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = v3dv_CmdPushConstants2,
    .CmdPushConstants2KHR = v3dv_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = v3dv_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = v3dv_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = v3dv_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = v3dv_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = v3dv_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = v3dv_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = v3dv_SetLatencySleepModeNV,
    .LatencySleepNV = v3dv_LatencySleepNV,
    .SetLatencyMarkerNV = v3dv_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = v3dv_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = v3dv_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = v3dv_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = v3dv_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = v3dv_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = v3dv_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = v3dv_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = v3dv_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = v3dv_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = v3dv_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = v3dv_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = v3dv_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = v3dv_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = v3dv_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = v3dv_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = v3dv_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = v3dv_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = v3dv_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = v3dv_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = v3dv_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = v3dv_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = v3dv_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = v3dv_CreateTensorARM,
    .DestroyTensorARM = v3dv_DestroyTensorARM,
    .CreateTensorViewARM = v3dv_CreateTensorViewARM,
    .DestroyTensorViewARM = v3dv_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = v3dv_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = v3dv_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = v3dv_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = v3dv_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = v3dv_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = v3dv_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = v3dv_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = v3dv_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = v3dv_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = v3dv_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = v3dv_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = v3dv_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = v3dv_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = v3dv_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = v3dv_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = v3dv_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = v3dv_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = v3dv_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = v3dv_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = v3dv_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = v3dv_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = v3dv_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = v3dv_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = v3dv_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = v3dv_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = v3dv_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = v3dv_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = v3dv_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = v3dv_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = v3dv_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = v3dv_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = v3dv_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = v3dv_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = v3dv_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = v3dv_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = v3dv_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = v3dv_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = v3dv_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = v3dv_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = v3dv_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = v3dv_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = v3dv_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = v3dv_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = v3dv_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = v3dv_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = v3dv_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = v3dv_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = v3dv_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = v3dv_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = v3dv_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = v3dv_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table ver42_device_entrypoints = {
    .GetDeviceProcAddr = ver42_GetDeviceProcAddr,
    .DestroyDevice = ver42_DestroyDevice,
    .GetDeviceQueue = ver42_GetDeviceQueue,
    .QueueSubmit = ver42_QueueSubmit,
    .QueueWaitIdle = ver42_QueueWaitIdle,
    .DeviceWaitIdle = ver42_DeviceWaitIdle,
    .AllocateMemory = ver42_AllocateMemory,
    .FreeMemory = ver42_FreeMemory,
    .MapMemory = ver42_MapMemory,
    .UnmapMemory = ver42_UnmapMemory,
    .FlushMappedMemoryRanges = ver42_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = ver42_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = ver42_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = ver42_GetBufferMemoryRequirements,
    .BindBufferMemory = ver42_BindBufferMemory,
    .GetImageMemoryRequirements = ver42_GetImageMemoryRequirements,
    .BindImageMemory = ver42_BindImageMemory,
    .GetImageSparseMemoryRequirements = ver42_GetImageSparseMemoryRequirements,
    .QueueBindSparse = ver42_QueueBindSparse,
    .CreateFence = ver42_CreateFence,
    .DestroyFence = ver42_DestroyFence,
    .ResetFences = ver42_ResetFences,
    .GetFenceStatus = ver42_GetFenceStatus,
    .WaitForFences = ver42_WaitForFences,
    .CreateSemaphore = ver42_CreateSemaphore,
    .DestroySemaphore = ver42_DestroySemaphore,
    .CreateEvent = ver42_CreateEvent,
    .DestroyEvent = ver42_DestroyEvent,
    .GetEventStatus = ver42_GetEventStatus,
    .SetEvent = ver42_SetEvent,
    .ResetEvent = ver42_ResetEvent,
    .CreateQueryPool = ver42_CreateQueryPool,
    .DestroyQueryPool = ver42_DestroyQueryPool,
    .GetQueryPoolResults = ver42_GetQueryPoolResults,
    .ResetQueryPool = ver42_ResetQueryPool,
    .ResetQueryPoolEXT = ver42_ResetQueryPoolEXT,
    .CreateBuffer = ver42_CreateBuffer,
    .DestroyBuffer = ver42_DestroyBuffer,
    .CreateBufferView = ver42_CreateBufferView,
    .DestroyBufferView = ver42_DestroyBufferView,
    .CreateImage = ver42_CreateImage,
    .DestroyImage = ver42_DestroyImage,
    .GetImageSubresourceLayout = ver42_GetImageSubresourceLayout,
    .CreateImageView = ver42_CreateImageView,
    .DestroyImageView = ver42_DestroyImageView,
    .CreateShaderModule = ver42_CreateShaderModule,
    .DestroyShaderModule = ver42_DestroyShaderModule,
    .CreatePipelineCache = ver42_CreatePipelineCache,
    .DestroyPipelineCache = ver42_DestroyPipelineCache,
    .GetPipelineCacheData = ver42_GetPipelineCacheData,
    .MergePipelineCaches = ver42_MergePipelineCaches,
    .CreatePipelineBinariesKHR = ver42_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = ver42_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = ver42_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = ver42_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = ver42_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = ver42_CreateGraphicsPipelines,
    .CreateComputePipelines = ver42_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = ver42_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = ver42_DestroyPipeline,
    .CreatePipelineLayout = ver42_CreatePipelineLayout,
    .DestroyPipelineLayout = ver42_DestroyPipelineLayout,
    .CreateSampler = ver42_CreateSampler,
    .DestroySampler = ver42_DestroySampler,
    .CreateDescriptorSetLayout = ver42_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = ver42_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = ver42_CreateDescriptorPool,
    .DestroyDescriptorPool = ver42_DestroyDescriptorPool,
    .ResetDescriptorPool = ver42_ResetDescriptorPool,
    .AllocateDescriptorSets = ver42_AllocateDescriptorSets,
    .FreeDescriptorSets = ver42_FreeDescriptorSets,
    .UpdateDescriptorSets = ver42_UpdateDescriptorSets,
    .CreateFramebuffer = ver42_CreateFramebuffer,
    .DestroyFramebuffer = ver42_DestroyFramebuffer,
    .CreateRenderPass = ver42_CreateRenderPass,
    .DestroyRenderPass = ver42_DestroyRenderPass,
    .GetRenderAreaGranularity = ver42_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = ver42_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = ver42_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = ver42_CreateCommandPool,
    .DestroyCommandPool = ver42_DestroyCommandPool,
    .ResetCommandPool = ver42_ResetCommandPool,
    .AllocateCommandBuffers = ver42_AllocateCommandBuffers,
    .FreeCommandBuffers = ver42_FreeCommandBuffers,
    .BeginCommandBuffer = ver42_BeginCommandBuffer,
    .EndCommandBuffer = ver42_EndCommandBuffer,
    .ResetCommandBuffer = ver42_ResetCommandBuffer,
    .CmdBindPipeline = ver42_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = ver42_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = ver42_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = ver42_CmdSetViewport,
    .CmdSetScissor = ver42_CmdSetScissor,
    .CmdSetLineWidth = ver42_CmdSetLineWidth,
    .CmdSetDepthBias = ver42_CmdSetDepthBias,
    .CmdSetBlendConstants = ver42_CmdSetBlendConstants,
    .CmdSetDepthBounds = ver42_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = ver42_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = ver42_CmdSetStencilWriteMask,
    .CmdSetStencilReference = ver42_CmdSetStencilReference,
    .CmdBindDescriptorSets = ver42_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = ver42_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = ver42_CmdBindVertexBuffers,
    .CmdDraw = ver42_CmdDraw,
    .CmdDrawIndexed = ver42_CmdDrawIndexed,
    .CmdDrawMultiEXT = ver42_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = ver42_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = ver42_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = ver42_CmdDrawIndexedIndirect,
    .CmdDispatch = ver42_CmdDispatch,
    .CmdDispatchIndirect = ver42_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = ver42_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = ver42_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = ver42_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = ver42_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = ver42_CmdCopyBuffer,
    .CmdCopyImage = ver42_CmdCopyImage,
    .CmdBlitImage = ver42_CmdBlitImage,
    .CmdCopyBufferToImage = ver42_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = ver42_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = ver42_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = ver42_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = ver42_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = ver42_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = ver42_CmdUpdateBuffer,
    .CmdFillBuffer = ver42_CmdFillBuffer,
    .CmdClearColorImage = ver42_CmdClearColorImage,
    .CmdClearDepthStencilImage = ver42_CmdClearDepthStencilImage,
    .CmdClearAttachments = ver42_CmdClearAttachments,
    .CmdResolveImage = ver42_CmdResolveImage,
    .CmdSetEvent = ver42_CmdSetEvent,
    .CmdResetEvent = ver42_CmdResetEvent,
    .CmdWaitEvents = ver42_CmdWaitEvents,
    .CmdPipelineBarrier = ver42_CmdPipelineBarrier,
    .CmdBeginQuery = ver42_CmdBeginQuery,
    .CmdEndQuery = ver42_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = ver42_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = ver42_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = ver42_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = ver42_CmdResetQueryPool,
    .CmdWriteTimestamp = ver42_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = ver42_CmdCopyQueryPoolResults,
    .CmdPushConstants = ver42_CmdPushConstants,
    .CmdBeginRenderPass = ver42_CmdBeginRenderPass,
    .CmdNextSubpass = ver42_CmdNextSubpass,
    .CmdEndRenderPass = ver42_CmdEndRenderPass,
    .CmdExecuteCommands = ver42_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = ver42_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = ver42_CreateSwapchainKHR,
    .DestroySwapchainKHR = ver42_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = ver42_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = ver42_AcquireNextImageKHR,
    .QueuePresentKHR = ver42_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = ver42_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = ver42_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = ver42_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = ver42_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = ver42_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = ver42_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = ver42_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = ver42_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = ver42_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = ver42_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = ver42_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = ver42_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = ver42_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = ver42_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = ver42_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = ver42_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = ver42_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = ver42_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = ver42_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = ver42_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = ver42_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = ver42_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = ver42_CmdPushDescriptorSetKHR,
    .TrimCommandPool = ver42_TrimCommandPool,
    .TrimCommandPoolKHR = ver42_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = ver42_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = ver42_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = ver42_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = ver42_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = ver42_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = ver42_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = ver42_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = ver42_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = ver42_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = ver42_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = ver42_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = ver42_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = ver42_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = ver42_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = ver42_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = ver42_GetFenceFdKHR,
    .ImportFenceFdKHR = ver42_ImportFenceFdKHR,
    .DisplayPowerControlEXT = ver42_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = ver42_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = ver42_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = ver42_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = ver42_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = ver42_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = ver42_BindBufferMemory2,
    .BindBufferMemory2KHR = ver42_BindBufferMemory2KHR,
    .BindImageMemory2 = ver42_BindImageMemory2,
    .BindImageMemory2KHR = ver42_BindImageMemory2KHR,
    .CmdSetDeviceMask = ver42_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = ver42_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = ver42_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = ver42_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = ver42_AcquireNextImage2KHR,
    .CmdDispatchBase = ver42_CmdDispatchBase,
    .CmdDispatchBaseKHR = ver42_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = ver42_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = ver42_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = ver42_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = ver42_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = ver42_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = ver42_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = ver42_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = ver42_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = ver42_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = ver42_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = ver42_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = ver42_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = ver42_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = ver42_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = ver42_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = ver42_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = ver42_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = ver42_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = ver42_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = ver42_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = ver42_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = ver42_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = ver42_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = ver42_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = ver42_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = ver42_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = ver42_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = ver42_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = ver42_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = ver42_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = ver42_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = ver42_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = ver42_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = ver42_GetDeviceQueue2,
    .CreateValidationCacheEXT = ver42_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = ver42_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = ver42_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = ver42_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = ver42_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = ver42_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = ver42_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = ver42_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = ver42_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = ver42_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = ver42_GetShaderInfoAMD,
    .SetLocalDimmingAMD = ver42_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = ver42_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = ver42_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = ver42_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = ver42_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = ver42_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = ver42_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = ver42_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = ver42_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = ver42_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = ver42_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = ver42_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = ver42_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = ver42_CreateRenderPass2,
    .CreateRenderPass2KHR = ver42_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = ver42_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = ver42_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = ver42_CmdNextSubpass2,
    .CmdNextSubpass2KHR = ver42_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = ver42_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = ver42_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = ver42_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = ver42_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = ver42_WaitSemaphores,
    .WaitSemaphoresKHR = ver42_WaitSemaphoresKHR,
    .SignalSemaphore = ver42_SignalSemaphore,
    .SignalSemaphoreKHR = ver42_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = ver42_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = ver42_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = ver42_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = ver42_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = ver42_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = ver42_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = ver42_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = ver42_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = ver42_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = ver42_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = ver42_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = ver42_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = ver42_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = ver42_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = ver42_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = ver42_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = ver42_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = ver42_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = ver42_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = ver42_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = ver42_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = ver42_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = ver42_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = ver42_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = ver42_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = ver42_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = ver42_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = ver42_CompileDeferredNV,
    .CreateAccelerationStructureNV = ver42_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = ver42_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = ver42_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = ver42_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = ver42_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = ver42_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = ver42_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = ver42_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = ver42_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = ver42_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = ver42_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = ver42_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = ver42_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = ver42_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = ver42_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = ver42_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = ver42_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = ver42_CmdTraceRaysKHR,
    .CmdTraceRaysNV = ver42_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = ver42_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = ver42_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = ver42_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = ver42_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = ver42_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = ver42_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = ver42_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = ver42_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = ver42_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = ver42_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = ver42_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = ver42_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = ver42_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = ver42_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = ver42_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = ver42_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = ver42_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = ver42_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = ver42_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = ver42_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = ver42_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = ver42_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = ver42_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = ver42_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = ver42_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = ver42_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = ver42_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = ver42_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = ver42_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = ver42_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = ver42_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = ver42_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = ver42_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = ver42_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = ver42_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = ver42_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = ver42_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = ver42_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = ver42_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = ver42_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = ver42_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = ver42_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = ver42_CmdSetLineStipple,
    .CmdSetLineStippleKHR = ver42_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = ver42_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = ver42_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = ver42_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = ver42_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = ver42_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = ver42_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = ver42_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = ver42_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = ver42_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = ver42_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = ver42_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = ver42_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = ver42_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = ver42_AntiLagUpdateAMD,
    .CmdSetCullMode = ver42_CmdSetCullMode,
    .CmdSetCullModeEXT = ver42_CmdSetCullModeEXT,
    .CmdSetFrontFace = ver42_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = ver42_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = ver42_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = ver42_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = ver42_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = ver42_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = ver42_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = ver42_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = ver42_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = ver42_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = ver42_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = ver42_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = ver42_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = ver42_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = ver42_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = ver42_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = ver42_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = ver42_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = ver42_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = ver42_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = ver42_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = ver42_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = ver42_CmdSetStencilOp,
    .CmdSetStencilOpEXT = ver42_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = ver42_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = ver42_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = ver42_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = ver42_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = ver42_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = ver42_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = ver42_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = ver42_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = ver42_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = ver42_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = ver42_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = ver42_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = ver42_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = ver42_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = ver42_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = ver42_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = ver42_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = ver42_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = ver42_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = ver42_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = ver42_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = ver42_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = ver42_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = ver42_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = ver42_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = ver42_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = ver42_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = ver42_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = ver42_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = ver42_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = ver42_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = ver42_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = ver42_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = ver42_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = ver42_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = ver42_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = ver42_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = ver42_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = ver42_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = ver42_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = ver42_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = ver42_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = ver42_DestroyPrivateDataSlotEXT,
    .SetPrivateData = ver42_SetPrivateData,
    .SetPrivateDataEXT = ver42_SetPrivateDataEXT,
    .GetPrivateData = ver42_GetPrivateData,
    .GetPrivateDataEXT = ver42_GetPrivateDataEXT,
    .CmdCopyBuffer2 = ver42_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = ver42_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = ver42_CmdCopyImage2,
    .CmdCopyImage2KHR = ver42_CmdCopyImage2KHR,
    .CmdBlitImage2 = ver42_CmdBlitImage2,
    .CmdBlitImage2KHR = ver42_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = ver42_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = ver42_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = ver42_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = ver42_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = ver42_CmdResolveImage2,
    .CmdResolveImage2KHR = ver42_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = ver42_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = ver42_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = ver42_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = ver42_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = ver42_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = ver42_CmdSetEvent2,
    .CmdSetEvent2KHR = ver42_CmdSetEvent2KHR,
    .CmdResetEvent2 = ver42_CmdResetEvent2,
    .CmdResetEvent2KHR = ver42_CmdResetEvent2KHR,
    .CmdWaitEvents2 = ver42_CmdWaitEvents2,
    .CmdWaitEvents2KHR = ver42_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = ver42_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = ver42_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = ver42_QueueSubmit2,
    .QueueSubmit2KHR = ver42_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = ver42_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = ver42_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = ver42_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = ver42_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = ver42_CopyMemoryToImage,
    .CopyMemoryToImageEXT = ver42_CopyMemoryToImageEXT,
    .CopyImageToMemory = ver42_CopyImageToMemory,
    .CopyImageToMemoryEXT = ver42_CopyImageToMemoryEXT,
    .CopyImageToImage = ver42_CopyImageToImage,
    .CopyImageToImageEXT = ver42_CopyImageToImageEXT,
    .TransitionImageLayout = ver42_TransitionImageLayout,
    .TransitionImageLayoutEXT = ver42_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = ver42_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = ver42_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = ver42_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = ver42_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = ver42_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = ver42_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = ver42_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = ver42_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = ver42_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = ver42_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = ver42_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = ver42_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = ver42_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = ver42_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = ver42_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = ver42_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = ver42_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = ver42_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = ver42_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = ver42_CreateCuModuleNVX,
    .CreateCuFunctionNVX = ver42_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = ver42_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = ver42_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = ver42_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = ver42_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = ver42_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = ver42_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = ver42_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = ver42_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = ver42_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = ver42_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = ver42_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = ver42_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = ver42_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = ver42_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = ver42_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = ver42_WaitForPresent2KHR,
    .WaitForPresentKHR = ver42_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = ver42_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = ver42_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = ver42_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = ver42_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = ver42_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = ver42_CmdBeginRendering,
    .CmdBeginRenderingKHR = ver42_CmdBeginRenderingKHR,
    .CmdEndRendering = ver42_CmdEndRendering,
    .CmdEndRendering2KHR = ver42_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = ver42_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = ver42_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = ver42_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = ver42_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = ver42_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = ver42_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = ver42_BuildMicromapsEXT,
    .DestroyMicromapEXT = ver42_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = ver42_CmdCopyMicromapEXT,
    .CopyMicromapEXT = ver42_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = ver42_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = ver42_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = ver42_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = ver42_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = ver42_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = ver42_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = ver42_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = ver42_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = ver42_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = ver42_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = ver42_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = ver42_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = ver42_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = ver42_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = ver42_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = ver42_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = ver42_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = ver42_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = ver42_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = ver42_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = ver42_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = ver42_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = ver42_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = ver42_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = ver42_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = ver42_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = ver42_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = ver42_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = ver42_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = ver42_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = ver42_MapMemory2,
    .MapMemory2KHR = ver42_MapMemory2KHR,
    .UnmapMemory2 = ver42_UnmapMemory2,
    .UnmapMemory2KHR = ver42_UnmapMemory2KHR,
    .CreateShadersEXT = ver42_CreateShadersEXT,
    .DestroyShaderEXT = ver42_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = ver42_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = ver42_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = ver42_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = ver42_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = ver42_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = ver42_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = ver42_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = ver42_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = ver42_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = ver42_CmdPushConstants2,
    .CmdPushConstants2KHR = ver42_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = ver42_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = ver42_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = ver42_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = ver42_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = ver42_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = ver42_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = ver42_SetLatencySleepModeNV,
    .LatencySleepNV = ver42_LatencySleepNV,
    .SetLatencyMarkerNV = ver42_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = ver42_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = ver42_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = ver42_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = ver42_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = ver42_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = ver42_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = ver42_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = ver42_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = ver42_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = ver42_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = ver42_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = ver42_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = ver42_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = ver42_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = ver42_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = ver42_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = ver42_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = ver42_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = ver42_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = ver42_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = ver42_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = ver42_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = ver42_CreateTensorARM,
    .DestroyTensorARM = ver42_DestroyTensorARM,
    .CreateTensorViewARM = ver42_CreateTensorViewARM,
    .DestroyTensorViewARM = ver42_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = ver42_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = ver42_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = ver42_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = ver42_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = ver42_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = ver42_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = ver42_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = ver42_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = ver42_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = ver42_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = ver42_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = ver42_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = ver42_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = ver42_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = ver42_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = ver42_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = ver42_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = ver42_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = ver42_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = ver42_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = ver42_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = ver42_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = ver42_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = ver42_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = ver42_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = ver42_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = ver42_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = ver42_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = ver42_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = ver42_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = ver42_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = ver42_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = ver42_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = ver42_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = ver42_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = ver42_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = ver42_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = ver42_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = ver42_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = ver42_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = ver42_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = ver42_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = ver42_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = ver42_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = ver42_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = ver42_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = ver42_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = ver42_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = ver42_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = ver42_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = ver42_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table ver71_device_entrypoints = {
    .GetDeviceProcAddr = ver71_GetDeviceProcAddr,
    .DestroyDevice = ver71_DestroyDevice,
    .GetDeviceQueue = ver71_GetDeviceQueue,
    .QueueSubmit = ver71_QueueSubmit,
    .QueueWaitIdle = ver71_QueueWaitIdle,
    .DeviceWaitIdle = ver71_DeviceWaitIdle,
    .AllocateMemory = ver71_AllocateMemory,
    .FreeMemory = ver71_FreeMemory,
    .MapMemory = ver71_MapMemory,
    .UnmapMemory = ver71_UnmapMemory,
    .FlushMappedMemoryRanges = ver71_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = ver71_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = ver71_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = ver71_GetBufferMemoryRequirements,
    .BindBufferMemory = ver71_BindBufferMemory,
    .GetImageMemoryRequirements = ver71_GetImageMemoryRequirements,
    .BindImageMemory = ver71_BindImageMemory,
    .GetImageSparseMemoryRequirements = ver71_GetImageSparseMemoryRequirements,
    .QueueBindSparse = ver71_QueueBindSparse,
    .CreateFence = ver71_CreateFence,
    .DestroyFence = ver71_DestroyFence,
    .ResetFences = ver71_ResetFences,
    .GetFenceStatus = ver71_GetFenceStatus,
    .WaitForFences = ver71_WaitForFences,
    .CreateSemaphore = ver71_CreateSemaphore,
    .DestroySemaphore = ver71_DestroySemaphore,
    .CreateEvent = ver71_CreateEvent,
    .DestroyEvent = ver71_DestroyEvent,
    .GetEventStatus = ver71_GetEventStatus,
    .SetEvent = ver71_SetEvent,
    .ResetEvent = ver71_ResetEvent,
    .CreateQueryPool = ver71_CreateQueryPool,
    .DestroyQueryPool = ver71_DestroyQueryPool,
    .GetQueryPoolResults = ver71_GetQueryPoolResults,
    .ResetQueryPool = ver71_ResetQueryPool,
    .ResetQueryPoolEXT = ver71_ResetQueryPoolEXT,
    .CreateBuffer = ver71_CreateBuffer,
    .DestroyBuffer = ver71_DestroyBuffer,
    .CreateBufferView = ver71_CreateBufferView,
    .DestroyBufferView = ver71_DestroyBufferView,
    .CreateImage = ver71_CreateImage,
    .DestroyImage = ver71_DestroyImage,
    .GetImageSubresourceLayout = ver71_GetImageSubresourceLayout,
    .CreateImageView = ver71_CreateImageView,
    .DestroyImageView = ver71_DestroyImageView,
    .CreateShaderModule = ver71_CreateShaderModule,
    .DestroyShaderModule = ver71_DestroyShaderModule,
    .CreatePipelineCache = ver71_CreatePipelineCache,
    .DestroyPipelineCache = ver71_DestroyPipelineCache,
    .GetPipelineCacheData = ver71_GetPipelineCacheData,
    .MergePipelineCaches = ver71_MergePipelineCaches,
    .CreatePipelineBinariesKHR = ver71_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = ver71_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = ver71_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = ver71_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = ver71_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = ver71_CreateGraphicsPipelines,
    .CreateComputePipelines = ver71_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = ver71_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = ver71_DestroyPipeline,
    .CreatePipelineLayout = ver71_CreatePipelineLayout,
    .DestroyPipelineLayout = ver71_DestroyPipelineLayout,
    .CreateSampler = ver71_CreateSampler,
    .DestroySampler = ver71_DestroySampler,
    .CreateDescriptorSetLayout = ver71_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = ver71_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = ver71_CreateDescriptorPool,
    .DestroyDescriptorPool = ver71_DestroyDescriptorPool,
    .ResetDescriptorPool = ver71_ResetDescriptorPool,
    .AllocateDescriptorSets = ver71_AllocateDescriptorSets,
    .FreeDescriptorSets = ver71_FreeDescriptorSets,
    .UpdateDescriptorSets = ver71_UpdateDescriptorSets,
    .CreateFramebuffer = ver71_CreateFramebuffer,
    .DestroyFramebuffer = ver71_DestroyFramebuffer,
    .CreateRenderPass = ver71_CreateRenderPass,
    .DestroyRenderPass = ver71_DestroyRenderPass,
    .GetRenderAreaGranularity = ver71_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = ver71_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = ver71_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = ver71_CreateCommandPool,
    .DestroyCommandPool = ver71_DestroyCommandPool,
    .ResetCommandPool = ver71_ResetCommandPool,
    .AllocateCommandBuffers = ver71_AllocateCommandBuffers,
    .FreeCommandBuffers = ver71_FreeCommandBuffers,
    .BeginCommandBuffer = ver71_BeginCommandBuffer,
    .EndCommandBuffer = ver71_EndCommandBuffer,
    .ResetCommandBuffer = ver71_ResetCommandBuffer,
    .CmdBindPipeline = ver71_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = ver71_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = ver71_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = ver71_CmdSetViewport,
    .CmdSetScissor = ver71_CmdSetScissor,
    .CmdSetLineWidth = ver71_CmdSetLineWidth,
    .CmdSetDepthBias = ver71_CmdSetDepthBias,
    .CmdSetBlendConstants = ver71_CmdSetBlendConstants,
    .CmdSetDepthBounds = ver71_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = ver71_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = ver71_CmdSetStencilWriteMask,
    .CmdSetStencilReference = ver71_CmdSetStencilReference,
    .CmdBindDescriptorSets = ver71_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = ver71_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = ver71_CmdBindVertexBuffers,
    .CmdDraw = ver71_CmdDraw,
    .CmdDrawIndexed = ver71_CmdDrawIndexed,
    .CmdDrawMultiEXT = ver71_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = ver71_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = ver71_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = ver71_CmdDrawIndexedIndirect,
    .CmdDispatch = ver71_CmdDispatch,
    .CmdDispatchIndirect = ver71_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = ver71_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = ver71_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = ver71_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = ver71_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = ver71_CmdCopyBuffer,
    .CmdCopyImage = ver71_CmdCopyImage,
    .CmdBlitImage = ver71_CmdBlitImage,
    .CmdCopyBufferToImage = ver71_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = ver71_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = ver71_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = ver71_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = ver71_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = ver71_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = ver71_CmdUpdateBuffer,
    .CmdFillBuffer = ver71_CmdFillBuffer,
    .CmdClearColorImage = ver71_CmdClearColorImage,
    .CmdClearDepthStencilImage = ver71_CmdClearDepthStencilImage,
    .CmdClearAttachments = ver71_CmdClearAttachments,
    .CmdResolveImage = ver71_CmdResolveImage,
    .CmdSetEvent = ver71_CmdSetEvent,
    .CmdResetEvent = ver71_CmdResetEvent,
    .CmdWaitEvents = ver71_CmdWaitEvents,
    .CmdPipelineBarrier = ver71_CmdPipelineBarrier,
    .CmdBeginQuery = ver71_CmdBeginQuery,
    .CmdEndQuery = ver71_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = ver71_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = ver71_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = ver71_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = ver71_CmdResetQueryPool,
    .CmdWriteTimestamp = ver71_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = ver71_CmdCopyQueryPoolResults,
    .CmdPushConstants = ver71_CmdPushConstants,
    .CmdBeginRenderPass = ver71_CmdBeginRenderPass,
    .CmdNextSubpass = ver71_CmdNextSubpass,
    .CmdEndRenderPass = ver71_CmdEndRenderPass,
    .CmdExecuteCommands = ver71_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = ver71_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = ver71_CreateSwapchainKHR,
    .DestroySwapchainKHR = ver71_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = ver71_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = ver71_AcquireNextImageKHR,
    .QueuePresentKHR = ver71_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = ver71_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = ver71_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = ver71_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = ver71_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = ver71_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = ver71_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = ver71_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = ver71_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = ver71_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = ver71_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = ver71_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = ver71_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = ver71_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = ver71_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = ver71_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = ver71_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = ver71_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = ver71_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = ver71_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = ver71_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = ver71_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = ver71_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = ver71_CmdPushDescriptorSetKHR,
    .TrimCommandPool = ver71_TrimCommandPool,
    .TrimCommandPoolKHR = ver71_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = ver71_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = ver71_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = ver71_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = ver71_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = ver71_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = ver71_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = ver71_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = ver71_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = ver71_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = ver71_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = ver71_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = ver71_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = ver71_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = ver71_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = ver71_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = ver71_GetFenceFdKHR,
    .ImportFenceFdKHR = ver71_ImportFenceFdKHR,
    .DisplayPowerControlEXT = ver71_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = ver71_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = ver71_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = ver71_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = ver71_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = ver71_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = ver71_BindBufferMemory2,
    .BindBufferMemory2KHR = ver71_BindBufferMemory2KHR,
    .BindImageMemory2 = ver71_BindImageMemory2,
    .BindImageMemory2KHR = ver71_BindImageMemory2KHR,
    .CmdSetDeviceMask = ver71_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = ver71_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = ver71_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = ver71_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = ver71_AcquireNextImage2KHR,
    .CmdDispatchBase = ver71_CmdDispatchBase,
    .CmdDispatchBaseKHR = ver71_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = ver71_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = ver71_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = ver71_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = ver71_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = ver71_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = ver71_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = ver71_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = ver71_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = ver71_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = ver71_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = ver71_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = ver71_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = ver71_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = ver71_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = ver71_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = ver71_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = ver71_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = ver71_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = ver71_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = ver71_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = ver71_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = ver71_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = ver71_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = ver71_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = ver71_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = ver71_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = ver71_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = ver71_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = ver71_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = ver71_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = ver71_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = ver71_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = ver71_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = ver71_GetDeviceQueue2,
    .CreateValidationCacheEXT = ver71_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = ver71_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = ver71_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = ver71_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = ver71_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = ver71_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = ver71_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = ver71_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = ver71_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = ver71_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = ver71_GetShaderInfoAMD,
    .SetLocalDimmingAMD = ver71_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = ver71_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = ver71_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = ver71_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = ver71_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = ver71_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = ver71_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = ver71_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = ver71_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = ver71_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = ver71_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = ver71_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = ver71_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = ver71_CreateRenderPass2,
    .CreateRenderPass2KHR = ver71_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = ver71_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = ver71_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = ver71_CmdNextSubpass2,
    .CmdNextSubpass2KHR = ver71_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = ver71_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = ver71_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = ver71_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = ver71_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = ver71_WaitSemaphores,
    .WaitSemaphoresKHR = ver71_WaitSemaphoresKHR,
    .SignalSemaphore = ver71_SignalSemaphore,
    .SignalSemaphoreKHR = ver71_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = ver71_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = ver71_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = ver71_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = ver71_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = ver71_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = ver71_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = ver71_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = ver71_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = ver71_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = ver71_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = ver71_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = ver71_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = ver71_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = ver71_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = ver71_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = ver71_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = ver71_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = ver71_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = ver71_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = ver71_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = ver71_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = ver71_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = ver71_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = ver71_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = ver71_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = ver71_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = ver71_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = ver71_CompileDeferredNV,
    .CreateAccelerationStructureNV = ver71_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = ver71_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = ver71_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = ver71_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = ver71_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = ver71_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = ver71_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = ver71_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = ver71_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = ver71_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = ver71_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = ver71_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = ver71_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = ver71_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = ver71_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = ver71_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = ver71_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = ver71_CmdTraceRaysKHR,
    .CmdTraceRaysNV = ver71_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = ver71_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = ver71_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = ver71_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = ver71_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = ver71_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = ver71_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = ver71_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = ver71_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = ver71_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = ver71_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = ver71_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = ver71_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = ver71_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = ver71_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = ver71_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = ver71_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = ver71_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = ver71_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = ver71_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = ver71_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = ver71_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = ver71_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = ver71_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = ver71_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = ver71_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = ver71_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = ver71_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = ver71_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = ver71_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = ver71_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = ver71_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = ver71_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = ver71_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = ver71_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = ver71_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = ver71_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = ver71_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = ver71_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = ver71_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = ver71_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = ver71_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = ver71_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = ver71_CmdSetLineStipple,
    .CmdSetLineStippleKHR = ver71_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = ver71_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = ver71_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = ver71_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = ver71_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = ver71_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = ver71_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = ver71_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = ver71_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = ver71_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = ver71_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = ver71_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = ver71_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = ver71_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = ver71_AntiLagUpdateAMD,
    .CmdSetCullMode = ver71_CmdSetCullMode,
    .CmdSetCullModeEXT = ver71_CmdSetCullModeEXT,
    .CmdSetFrontFace = ver71_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = ver71_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = ver71_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = ver71_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = ver71_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = ver71_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = ver71_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = ver71_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = ver71_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = ver71_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = ver71_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = ver71_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = ver71_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = ver71_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = ver71_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = ver71_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = ver71_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = ver71_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = ver71_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = ver71_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = ver71_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = ver71_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = ver71_CmdSetStencilOp,
    .CmdSetStencilOpEXT = ver71_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = ver71_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = ver71_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = ver71_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = ver71_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = ver71_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = ver71_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = ver71_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = ver71_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = ver71_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = ver71_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = ver71_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = ver71_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = ver71_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = ver71_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = ver71_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = ver71_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = ver71_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = ver71_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = ver71_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = ver71_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = ver71_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = ver71_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = ver71_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = ver71_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = ver71_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = ver71_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = ver71_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = ver71_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = ver71_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = ver71_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = ver71_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = ver71_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = ver71_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = ver71_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = ver71_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = ver71_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = ver71_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = ver71_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = ver71_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = ver71_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = ver71_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = ver71_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = ver71_DestroyPrivateDataSlotEXT,
    .SetPrivateData = ver71_SetPrivateData,
    .SetPrivateDataEXT = ver71_SetPrivateDataEXT,
    .GetPrivateData = ver71_GetPrivateData,
    .GetPrivateDataEXT = ver71_GetPrivateDataEXT,
    .CmdCopyBuffer2 = ver71_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = ver71_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = ver71_CmdCopyImage2,
    .CmdCopyImage2KHR = ver71_CmdCopyImage2KHR,
    .CmdBlitImage2 = ver71_CmdBlitImage2,
    .CmdBlitImage2KHR = ver71_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = ver71_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = ver71_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = ver71_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = ver71_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = ver71_CmdResolveImage2,
    .CmdResolveImage2KHR = ver71_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = ver71_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = ver71_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = ver71_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = ver71_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = ver71_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = ver71_CmdSetEvent2,
    .CmdSetEvent2KHR = ver71_CmdSetEvent2KHR,
    .CmdResetEvent2 = ver71_CmdResetEvent2,
    .CmdResetEvent2KHR = ver71_CmdResetEvent2KHR,
    .CmdWaitEvents2 = ver71_CmdWaitEvents2,
    .CmdWaitEvents2KHR = ver71_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = ver71_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = ver71_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = ver71_QueueSubmit2,
    .QueueSubmit2KHR = ver71_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = ver71_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = ver71_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = ver71_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = ver71_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = ver71_CopyMemoryToImage,
    .CopyMemoryToImageEXT = ver71_CopyMemoryToImageEXT,
    .CopyImageToMemory = ver71_CopyImageToMemory,
    .CopyImageToMemoryEXT = ver71_CopyImageToMemoryEXT,
    .CopyImageToImage = ver71_CopyImageToImage,
    .CopyImageToImageEXT = ver71_CopyImageToImageEXT,
    .TransitionImageLayout = ver71_TransitionImageLayout,
    .TransitionImageLayoutEXT = ver71_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = ver71_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = ver71_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = ver71_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = ver71_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = ver71_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = ver71_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = ver71_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = ver71_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = ver71_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = ver71_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = ver71_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = ver71_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = ver71_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = ver71_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = ver71_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = ver71_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = ver71_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = ver71_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = ver71_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = ver71_CreateCuModuleNVX,
    .CreateCuFunctionNVX = ver71_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = ver71_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = ver71_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = ver71_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = ver71_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = ver71_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = ver71_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = ver71_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = ver71_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = ver71_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = ver71_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = ver71_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = ver71_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = ver71_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = ver71_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = ver71_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = ver71_WaitForPresent2KHR,
    .WaitForPresentKHR = ver71_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = ver71_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = ver71_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = ver71_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = ver71_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = ver71_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = ver71_CmdBeginRendering,
    .CmdBeginRenderingKHR = ver71_CmdBeginRenderingKHR,
    .CmdEndRendering = ver71_CmdEndRendering,
    .CmdEndRendering2KHR = ver71_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = ver71_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = ver71_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = ver71_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = ver71_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = ver71_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = ver71_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = ver71_BuildMicromapsEXT,
    .DestroyMicromapEXT = ver71_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = ver71_CmdCopyMicromapEXT,
    .CopyMicromapEXT = ver71_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = ver71_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = ver71_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = ver71_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = ver71_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = ver71_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = ver71_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = ver71_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = ver71_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = ver71_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = ver71_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = ver71_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = ver71_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = ver71_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = ver71_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = ver71_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = ver71_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = ver71_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = ver71_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = ver71_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = ver71_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = ver71_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = ver71_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = ver71_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = ver71_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = ver71_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = ver71_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = ver71_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = ver71_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = ver71_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = ver71_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = ver71_MapMemory2,
    .MapMemory2KHR = ver71_MapMemory2KHR,
    .UnmapMemory2 = ver71_UnmapMemory2,
    .UnmapMemory2KHR = ver71_UnmapMemory2KHR,
    .CreateShadersEXT = ver71_CreateShadersEXT,
    .DestroyShaderEXT = ver71_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = ver71_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = ver71_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = ver71_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = ver71_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = ver71_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = ver71_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = ver71_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = ver71_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = ver71_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = ver71_CmdPushConstants2,
    .CmdPushConstants2KHR = ver71_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = ver71_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = ver71_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = ver71_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = ver71_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = ver71_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = ver71_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = ver71_SetLatencySleepModeNV,
    .LatencySleepNV = ver71_LatencySleepNV,
    .SetLatencyMarkerNV = ver71_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = ver71_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = ver71_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = ver71_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = ver71_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = ver71_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = ver71_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = ver71_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = ver71_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = ver71_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = ver71_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = ver71_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = ver71_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = ver71_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = ver71_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = ver71_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = ver71_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = ver71_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = ver71_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = ver71_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = ver71_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = ver71_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = ver71_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = ver71_CreateTensorARM,
    .DestroyTensorARM = ver71_DestroyTensorARM,
    .CreateTensorViewARM = ver71_CreateTensorViewARM,
    .DestroyTensorViewARM = ver71_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = ver71_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = ver71_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = ver71_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = ver71_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = ver71_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = ver71_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = ver71_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = ver71_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = ver71_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = ver71_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = ver71_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = ver71_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = ver71_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = ver71_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = ver71_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = ver71_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = ver71_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = ver71_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = ver71_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = ver71_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = ver71_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = ver71_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = ver71_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = ver71_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = ver71_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = ver71_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = ver71_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = ver71_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = ver71_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = ver71_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = ver71_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = ver71_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = ver71_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = ver71_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = ver71_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = ver71_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = ver71_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = ver71_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = ver71_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = ver71_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = ver71_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = ver71_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = ver71_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = ver71_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = ver71_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = ver71_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = ver71_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = ver71_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = ver71_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = ver71_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = ver71_CmdSetDispatchParametersARM,
};


