
/* Copyright © 2015-2021 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/* This file generated from vk_entrypoints_gen.py, don't edit directly. */

/* This file is the only place we rely on undefined symbols to fall back to
 * NULL. Other files use regular symbol declarations.
 * See also comments on VK_ENTRY_WEAK.
 */
#define VK_ENTRY_USE_WEAK 1
#include "vn_entrypoints.h"

/* Weak aliases for all potential implementations. These will resolve to
 * NULL if they're not defined, which lets the resolve_entrypoint() function
 * either pick the correct entry point.
 *
 * MSVC uses different decorated names for 32-bit versus 64-bit. Declare
 * all argument sizes for 32-bit because computing the actual size would be
 * difficult.
 */




#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateInstance@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateInstance=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateInstance=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyInstance@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyInstance=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyInstance=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDevices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumeratePhysicalDevices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumeratePhysicalDevices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetInstanceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetInstanceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetInstanceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceVersion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumerateInstanceVersion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumerateInstanceVersion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceLayerProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumerateInstanceLayerProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumerateInstanceLayerProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateInstanceExtensionProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumerateInstanceExtensionProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumerateInstanceExtensionProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAndroidSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateAndroidSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateAndroidSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSurfaceOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateSurfaceOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateSurfaceOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayPlaneSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDisplayPlaneSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDisplayPlaneSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroySurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroySurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_VI_NN
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateViSurfaceNN@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateViSurfaceNN=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateViSurfaceNN=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_VI_NN
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWaylandSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateWaylandSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateWaylandSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateUbmSurfaceSEC@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateUbmSurfaceSEC=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateUbmSurfaceSEC=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateWin32SurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateWin32SurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateWin32SurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXlibSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateXlibSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateXlibSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateXcbSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateXcbSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateXcbSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDirectFBSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDirectFBSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDirectFBSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImagePipeSurfaceFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateImagePipeSurfaceFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateImagePipeSurfaceFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_GGP
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateStreamDescriptorSurfaceGGP@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateStreamDescriptorSurfaceGGP=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateStreamDescriptorSurfaceGGP=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_GGP
#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateScreenSurfaceQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateScreenSurfaceQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateScreenSurfaceQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugReportCallbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDebugReportCallbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDebugReportCallbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugReportCallbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyDebugReportCallbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyDebugReportCallbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugReportMessageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DebugReportMessageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DebugReportMessageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroups@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumeratePhysicalDeviceGroups=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumeratePhysicalDeviceGroups=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceGroupsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumeratePhysicalDeviceGroupsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumeratePhysicalDeviceGroupsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_IOS_MVK
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIOSSurfaceMVK@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateIOSSurfaceMVK=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateIOSSurfaceMVK=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_IOS_MVK
#ifdef VK_USE_PLATFORM_MACOS_MVK
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMacOSSurfaceMVK@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateMacOSSurfaceMVK=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateMacOSSurfaceMVK=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_MACOS_MVK
#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMetalSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateMetalSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateMetalSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDebugUtilsMessengerEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDebugUtilsMessengerEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDebugUtilsMessengerEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDebugUtilsMessengerEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyDebugUtilsMessengerEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyDebugUtilsMessengerEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SubmitDebugUtilsMessageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SubmitDebugUtilsMessageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SubmitDebugUtilsMessageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateHeadlessSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateHeadlessSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateHeadlessSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetExternalComputeQueueDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetExternalComputeQueueDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetExternalComputeQueueDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_instance_entrypoint_table vn_instance_entrypoints = {
    .CreateInstance = vn_CreateInstance,
    .DestroyInstance = vn_DestroyInstance,
    .EnumeratePhysicalDevices = vn_EnumeratePhysicalDevices,
    .GetInstanceProcAddr = vn_GetInstanceProcAddr,
    .EnumerateInstanceVersion = vn_EnumerateInstanceVersion,
    .EnumerateInstanceLayerProperties = vn_EnumerateInstanceLayerProperties,
    .EnumerateInstanceExtensionProperties = vn_EnumerateInstanceExtensionProperties,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .CreateAndroidSurfaceKHR = vn_CreateAndroidSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateAndroidSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_OHOS
    .CreateSurfaceOHOS = vn_CreateSurfaceOHOS,
#elif defined(_MSC_VER)
    .CreateSurfaceOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .CreateDisplayPlaneSurfaceKHR = vn_CreateDisplayPlaneSurfaceKHR,
    .DestroySurfaceKHR = vn_DestroySurfaceKHR,
#ifdef VK_USE_PLATFORM_VI_NN
    .CreateViSurfaceNN = vn_CreateViSurfaceNN,
#elif defined(_MSC_VER)
    .CreateViSurfaceNN = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_VI_NN
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
    .CreateWaylandSurfaceKHR = vn_CreateWaylandSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateWaylandSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
    .CreateUbmSurfaceSEC = vn_CreateUbmSurfaceSEC,
#elif defined(_MSC_VER)
    .CreateUbmSurfaceSEC = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .CreateWin32SurfaceKHR = vn_CreateWin32SurfaceKHR,
#elif defined(_MSC_VER)
    .CreateWin32SurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
    .CreateXlibSurfaceKHR = vn_CreateXlibSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateXlibSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
    .CreateXcbSurfaceKHR = vn_CreateXcbSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateXcbSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
    .CreateDirectFBSurfaceEXT = vn_CreateDirectFBSurfaceEXT,
#elif defined(_MSC_VER)
    .CreateDirectFBSurfaceEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateImagePipeSurfaceFUCHSIA = vn_CreateImagePipeSurfaceFUCHSIA,
#elif defined(_MSC_VER)
    .CreateImagePipeSurfaceFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_GGP
    .CreateStreamDescriptorSurfaceGGP = vn_CreateStreamDescriptorSurfaceGGP,
#elif defined(_MSC_VER)
    .CreateStreamDescriptorSurfaceGGP = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_GGP
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .CreateScreenSurfaceQNX = vn_CreateScreenSurfaceQNX,
#elif defined(_MSC_VER)
    .CreateScreenSurfaceQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CreateDebugReportCallbackEXT = vn_CreateDebugReportCallbackEXT,
    .DestroyDebugReportCallbackEXT = vn_DestroyDebugReportCallbackEXT,
    .DebugReportMessageEXT = vn_DebugReportMessageEXT,
    .EnumeratePhysicalDeviceGroups = vn_EnumeratePhysicalDeviceGroups,
    .EnumeratePhysicalDeviceGroupsKHR = vn_EnumeratePhysicalDeviceGroupsKHR,
#ifdef VK_USE_PLATFORM_IOS_MVK
    .CreateIOSSurfaceMVK = vn_CreateIOSSurfaceMVK,
#elif defined(_MSC_VER)
    .CreateIOSSurfaceMVK = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_IOS_MVK
#ifdef VK_USE_PLATFORM_MACOS_MVK
    .CreateMacOSSurfaceMVK = vn_CreateMacOSSurfaceMVK,
#elif defined(_MSC_VER)
    .CreateMacOSSurfaceMVK = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_MACOS_MVK
#ifdef VK_USE_PLATFORM_METAL_EXT
    .CreateMetalSurfaceEXT = vn_CreateMetalSurfaceEXT,
#elif defined(_MSC_VER)
    .CreateMetalSurfaceEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CreateDebugUtilsMessengerEXT = vn_CreateDebugUtilsMessengerEXT,
    .DestroyDebugUtilsMessengerEXT = vn_DestroyDebugUtilsMessengerEXT,
    .SubmitDebugUtilsMessageEXT = vn_SubmitDebugUtilsMessageEXT,
    .CreateHeadlessSurfaceEXT = vn_CreateHeadlessSurfaceEXT,
    .GetExternalComputeQueueDataNV = vn_GetExternalComputeQueueDataNV,
};



#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceQueueFamilyProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceQueueFamilyProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceMemoryProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceMemoryProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceImageFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceImageFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceLayerProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumerateDeviceLayerProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumerateDeviceLayerProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumerateDeviceExtensionProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumerateDeviceExtensionProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumerateDeviceExtensionProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSparseImageFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSparseImageFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceDisplayPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceDisplayPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlanePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceDisplayPlanePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceDisplayPlanePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneSupportedDisplaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDisplayPlaneSupportedDisplaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDisplayPlaneSupportedDisplaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDisplayModePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDisplayModePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDisplayModeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDisplayModeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDisplayModeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDisplayPlaneCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDisplayPlaneCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSurfaceSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSurfaceSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSurfaceCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSurfaceCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormatsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSurfaceFormatsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSurfaceFormatsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WAYLAND_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWaylandPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceWaylandPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceWaylandPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceUbmPresentationSupportSEC@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceUbmPresentationSupportSEC=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceUbmPresentationSupportSEC=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceWin32PresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceWin32PresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceWin32PresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXlibPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceXlibPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceXlibPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceXcbPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceXcbPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceXcbPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDirectFBPresentationSupportEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceDirectFBPresentationSupportEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceDirectFBPresentationSupportEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceScreenPresentationSupportQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceScreenPresentationSupportQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceScreenPresentationSupportQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalImageFormatPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceExternalImageFormatPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceExternalImageFormatPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceFeatures2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceFeatures2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFeatures2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceFeatures2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceFeatures2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceImageFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceImageFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceImageFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceImageFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceImageFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceQueueFamilyProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceQueueFamilyProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceQueueFamilyProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceQueueFamilyProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceMemoryProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceMemoryProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMemoryProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceMemoryProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceMemoryProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSparseImageFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSparseImageFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSparseImageFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSparseImageFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSparseImageFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceExternalBufferProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceExternalBufferProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalBufferPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceExternalBufferPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceExternalBufferPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphoreProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceExternalSemaphoreProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceExternalSemaphoreProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFenceProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceExternalFenceProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceExternalFenceProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalFencePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceExternalFencePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceExternalFencePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ReleaseDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ReleaseDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireXlibDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AcquireXlibDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AcquireXlibDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRandROutputDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetRandROutputDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetRandROutputDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireWinrtDisplayNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AcquireWinrtDisplayNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AcquireWinrtDisplayNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetWinrtDisplayNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetWinrtDisplayNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetWinrtDisplayNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSurfaceCapabilities2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSurfaceCapabilities2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDevicePresentRectanglesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDevicePresentRectanglesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDevicePresentRectanglesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceMultisamplePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceMultisamplePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceMultisamplePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceCapabilities2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSurfaceCapabilities2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSurfaceCapabilities2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfaceFormats2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSurfaceFormats2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSurfaceFormats2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceDisplayProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceDisplayProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDisplayPlaneProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceDisplayPlaneProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceDisplayPlaneProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayModeProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDisplayModeProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDisplayModeProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDisplayPlaneCapabilities2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDisplayPlaneCapabilities2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDisplayPlaneCapabilities2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceToolProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceToolProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceToolPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceToolPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceToolPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceFragmentShadingRatesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceFragmentShadingRatesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceFragmentShadingRatesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceVideoCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceVideoCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoFormatPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceVideoFormatPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceVideoFormatPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireDrmDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AcquireDrmDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AcquireDrmDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDrmDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDrmDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDrmDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceOpticalFlowImageFormatsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceOpticalFlowImageFormatsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceOpticalFlowImageFormatsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceCooperativeVectorPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceCooperativeVectorPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceCooperativeVectorPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceExternalTensorPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceExternalTensorPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceExternalTensorPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceDescriptorSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceDescriptorSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceDescriptorSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_physical_device_entrypoint_table vn_physical_device_entrypoints = {
    .GetPhysicalDeviceProperties = vn_GetPhysicalDeviceProperties,
    .GetPhysicalDeviceQueueFamilyProperties = vn_GetPhysicalDeviceQueueFamilyProperties,
    .GetPhysicalDeviceMemoryProperties = vn_GetPhysicalDeviceMemoryProperties,
    .GetPhysicalDeviceFeatures = vn_GetPhysicalDeviceFeatures,
    .GetPhysicalDeviceFormatProperties = vn_GetPhysicalDeviceFormatProperties,
    .GetPhysicalDeviceImageFormatProperties = vn_GetPhysicalDeviceImageFormatProperties,
    .CreateDevice = vn_CreateDevice,
    .EnumerateDeviceLayerProperties = vn_EnumerateDeviceLayerProperties,
    .EnumerateDeviceExtensionProperties = vn_EnumerateDeviceExtensionProperties,
    .GetPhysicalDeviceSparseImageFormatProperties = vn_GetPhysicalDeviceSparseImageFormatProperties,
    .GetPhysicalDeviceDisplayPropertiesKHR = vn_GetPhysicalDeviceDisplayPropertiesKHR,
    .GetPhysicalDeviceDisplayPlanePropertiesKHR = vn_GetPhysicalDeviceDisplayPlanePropertiesKHR,
    .GetDisplayPlaneSupportedDisplaysKHR = vn_GetDisplayPlaneSupportedDisplaysKHR,
    .GetDisplayModePropertiesKHR = vn_GetDisplayModePropertiesKHR,
    .CreateDisplayModeKHR = vn_CreateDisplayModeKHR,
    .GetDisplayPlaneCapabilitiesKHR = vn_GetDisplayPlaneCapabilitiesKHR,
    .GetPhysicalDeviceSurfaceSupportKHR = vn_GetPhysicalDeviceSurfaceSupportKHR,
    .GetPhysicalDeviceSurfaceCapabilitiesKHR = vn_GetPhysicalDeviceSurfaceCapabilitiesKHR,
    .GetPhysicalDeviceSurfaceFormatsKHR = vn_GetPhysicalDeviceSurfaceFormatsKHR,
    .GetPhysicalDeviceSurfacePresentModesKHR = vn_GetPhysicalDeviceSurfacePresentModesKHR,
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
    .GetPhysicalDeviceWaylandPresentationSupportKHR = vn_GetPhysicalDeviceWaylandPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceWaylandPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
    .GetPhysicalDeviceUbmPresentationSupportSEC = vn_GetPhysicalDeviceUbmPresentationSupportSEC,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceUbmPresentationSupportSEC = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceWin32PresentationSupportKHR = vn_GetPhysicalDeviceWin32PresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceWin32PresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
    .GetPhysicalDeviceXlibPresentationSupportKHR = vn_GetPhysicalDeviceXlibPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceXlibPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
    .GetPhysicalDeviceXcbPresentationSupportKHR = vn_GetPhysicalDeviceXcbPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceXcbPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
    .GetPhysicalDeviceDirectFBPresentationSupportEXT = vn_GetPhysicalDeviceDirectFBPresentationSupportEXT,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceDirectFBPresentationSupportEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetPhysicalDeviceScreenPresentationSupportQNX = vn_GetPhysicalDeviceScreenPresentationSupportQNX,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceScreenPresentationSupportQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .GetPhysicalDeviceExternalImageFormatPropertiesNV = vn_GetPhysicalDeviceExternalImageFormatPropertiesNV,
    .GetPhysicalDeviceFeatures2 = vn_GetPhysicalDeviceFeatures2,
    .GetPhysicalDeviceFeatures2KHR = vn_GetPhysicalDeviceFeatures2KHR,
    .GetPhysicalDeviceProperties2 = vn_GetPhysicalDeviceProperties2,
    .GetPhysicalDeviceProperties2KHR = vn_GetPhysicalDeviceProperties2KHR,
    .GetPhysicalDeviceFormatProperties2 = vn_GetPhysicalDeviceFormatProperties2,
    .GetPhysicalDeviceFormatProperties2KHR = vn_GetPhysicalDeviceFormatProperties2KHR,
    .GetPhysicalDeviceImageFormatProperties2 = vn_GetPhysicalDeviceImageFormatProperties2,
    .GetPhysicalDeviceImageFormatProperties2KHR = vn_GetPhysicalDeviceImageFormatProperties2KHR,
    .GetPhysicalDeviceQueueFamilyProperties2 = vn_GetPhysicalDeviceQueueFamilyProperties2,
    .GetPhysicalDeviceQueueFamilyProperties2KHR = vn_GetPhysicalDeviceQueueFamilyProperties2KHR,
    .GetPhysicalDeviceMemoryProperties2 = vn_GetPhysicalDeviceMemoryProperties2,
    .GetPhysicalDeviceMemoryProperties2KHR = vn_GetPhysicalDeviceMemoryProperties2KHR,
    .GetPhysicalDeviceSparseImageFormatProperties2 = vn_GetPhysicalDeviceSparseImageFormatProperties2,
    .GetPhysicalDeviceSparseImageFormatProperties2KHR = vn_GetPhysicalDeviceSparseImageFormatProperties2KHR,
    .GetPhysicalDeviceExternalBufferProperties = vn_GetPhysicalDeviceExternalBufferProperties,
    .GetPhysicalDeviceExternalBufferPropertiesKHR = vn_GetPhysicalDeviceExternalBufferPropertiesKHR,
    .GetPhysicalDeviceExternalSemaphoreProperties = vn_GetPhysicalDeviceExternalSemaphoreProperties,
    .GetPhysicalDeviceExternalSemaphorePropertiesKHR = vn_GetPhysicalDeviceExternalSemaphorePropertiesKHR,
    .GetPhysicalDeviceExternalFenceProperties = vn_GetPhysicalDeviceExternalFenceProperties,
    .GetPhysicalDeviceExternalFencePropertiesKHR = vn_GetPhysicalDeviceExternalFencePropertiesKHR,
    .ReleaseDisplayEXT = vn_ReleaseDisplayEXT,
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
    .AcquireXlibDisplayEXT = vn_AcquireXlibDisplayEXT,
#elif defined(_MSC_VER)
    .AcquireXlibDisplayEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
    .GetRandROutputDisplayEXT = vn_GetRandROutputDisplayEXT,
#elif defined(_MSC_VER)
    .GetRandROutputDisplayEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireWinrtDisplayNV = vn_AcquireWinrtDisplayNV,
#elif defined(_MSC_VER)
    .AcquireWinrtDisplayNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetWinrtDisplayNV = vn_GetWinrtDisplayNV,
#elif defined(_MSC_VER)
    .GetWinrtDisplayNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceSurfaceCapabilities2EXT = vn_GetPhysicalDeviceSurfaceCapabilities2EXT,
    .GetPhysicalDevicePresentRectanglesKHR = vn_GetPhysicalDevicePresentRectanglesKHR,
    .GetPhysicalDeviceMultisamplePropertiesEXT = vn_GetPhysicalDeviceMultisamplePropertiesEXT,
    .GetPhysicalDeviceSurfaceCapabilities2KHR = vn_GetPhysicalDeviceSurfaceCapabilities2KHR,
    .GetPhysicalDeviceSurfaceFormats2KHR = vn_GetPhysicalDeviceSurfaceFormats2KHR,
    .GetPhysicalDeviceDisplayProperties2KHR = vn_GetPhysicalDeviceDisplayProperties2KHR,
    .GetPhysicalDeviceDisplayPlaneProperties2KHR = vn_GetPhysicalDeviceDisplayPlaneProperties2KHR,
    .GetDisplayModeProperties2KHR = vn_GetDisplayModeProperties2KHR,
    .GetDisplayPlaneCapabilities2KHR = vn_GetDisplayPlaneCapabilities2KHR,
    .GetPhysicalDeviceCalibrateableTimeDomainsKHR = vn_GetPhysicalDeviceCalibrateableTimeDomainsKHR,
    .GetPhysicalDeviceCalibrateableTimeDomainsEXT = vn_GetPhysicalDeviceCalibrateableTimeDomainsEXT,
    .GetPhysicalDeviceCooperativeMatrixPropertiesNV = vn_GetPhysicalDeviceCooperativeMatrixPropertiesNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceSurfacePresentModes2EXT = vn_GetPhysicalDeviceSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR = vn_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR,
    .GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR = vn_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR,
    .GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV = vn_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV,
    .GetPhysicalDeviceToolProperties = vn_GetPhysicalDeviceToolProperties,
    .GetPhysicalDeviceToolPropertiesEXT = vn_GetPhysicalDeviceToolPropertiesEXT,
    .GetPhysicalDeviceFragmentShadingRatesKHR = vn_GetPhysicalDeviceFragmentShadingRatesKHR,
    .GetPhysicalDeviceVideoCapabilitiesKHR = vn_GetPhysicalDeviceVideoCapabilitiesKHR,
    .GetPhysicalDeviceVideoFormatPropertiesKHR = vn_GetPhysicalDeviceVideoFormatPropertiesKHR,
    .GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR = vn_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR,
    .AcquireDrmDisplayEXT = vn_AcquireDrmDisplayEXT,
    .GetDrmDisplayEXT = vn_GetDrmDisplayEXT,
    .GetPhysicalDeviceOpticalFlowImageFormatsNV = vn_GetPhysicalDeviceOpticalFlowImageFormatsNV,
    .GetPhysicalDeviceCooperativeMatrixPropertiesKHR = vn_GetPhysicalDeviceCooperativeMatrixPropertiesKHR,
    .GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV = vn_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV,
    .GetPhysicalDeviceCooperativeVectorPropertiesNV = vn_GetPhysicalDeviceCooperativeVectorPropertiesNV,
    .EnumeratePhysicalDeviceShaderInstrumentationMetricsARM = vn_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM,
    .GetPhysicalDeviceExternalTensorPropertiesARM = vn_GetPhysicalDeviceExternalTensorPropertiesARM,
    .GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM = vn_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM,
    .GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM = vn_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM,
    .EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM = vn_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM,
    .GetPhysicalDeviceDescriptorSizeEXT = vn_GetPhysicalDeviceDescriptorSizeEXT,
    .GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM = vn_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM,
};



#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_vn_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:vn_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#vn_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_device_entrypoint_table vn_device_entrypoints = {
    .GetDeviceProcAddr = vn_GetDeviceProcAddr,
    .DestroyDevice = vn_DestroyDevice,
    .GetDeviceQueue = vn_GetDeviceQueue,
    .QueueSubmit = vn_QueueSubmit,
    .QueueWaitIdle = vn_QueueWaitIdle,
    .DeviceWaitIdle = vn_DeviceWaitIdle,
    .AllocateMemory = vn_AllocateMemory,
    .FreeMemory = vn_FreeMemory,
    .MapMemory = vn_MapMemory,
    .UnmapMemory = vn_UnmapMemory,
    .FlushMappedMemoryRanges = vn_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = vn_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = vn_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = vn_GetBufferMemoryRequirements,
    .BindBufferMemory = vn_BindBufferMemory,
    .GetImageMemoryRequirements = vn_GetImageMemoryRequirements,
    .BindImageMemory = vn_BindImageMemory,
    .GetImageSparseMemoryRequirements = vn_GetImageSparseMemoryRequirements,
    .QueueBindSparse = vn_QueueBindSparse,
    .CreateFence = vn_CreateFence,
    .DestroyFence = vn_DestroyFence,
    .ResetFences = vn_ResetFences,
    .GetFenceStatus = vn_GetFenceStatus,
    .WaitForFences = vn_WaitForFences,
    .CreateSemaphore = vn_CreateSemaphore,
    .DestroySemaphore = vn_DestroySemaphore,
    .CreateEvent = vn_CreateEvent,
    .DestroyEvent = vn_DestroyEvent,
    .GetEventStatus = vn_GetEventStatus,
    .SetEvent = vn_SetEvent,
    .ResetEvent = vn_ResetEvent,
    .CreateQueryPool = vn_CreateQueryPool,
    .DestroyQueryPool = vn_DestroyQueryPool,
    .GetQueryPoolResults = vn_GetQueryPoolResults,
    .ResetQueryPool = vn_ResetQueryPool,
    .ResetQueryPoolEXT = vn_ResetQueryPoolEXT,
    .CreateBuffer = vn_CreateBuffer,
    .DestroyBuffer = vn_DestroyBuffer,
    .CreateBufferView = vn_CreateBufferView,
    .DestroyBufferView = vn_DestroyBufferView,
    .CreateImage = vn_CreateImage,
    .DestroyImage = vn_DestroyImage,
    .GetImageSubresourceLayout = vn_GetImageSubresourceLayout,
    .CreateImageView = vn_CreateImageView,
    .DestroyImageView = vn_DestroyImageView,
    .CreateShaderModule = vn_CreateShaderModule,
    .DestroyShaderModule = vn_DestroyShaderModule,
    .CreatePipelineCache = vn_CreatePipelineCache,
    .DestroyPipelineCache = vn_DestroyPipelineCache,
    .GetPipelineCacheData = vn_GetPipelineCacheData,
    .MergePipelineCaches = vn_MergePipelineCaches,
    .CreatePipelineBinariesKHR = vn_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = vn_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = vn_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = vn_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = vn_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = vn_CreateGraphicsPipelines,
    .CreateComputePipelines = vn_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = vn_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = vn_DestroyPipeline,
    .CreatePipelineLayout = vn_CreatePipelineLayout,
    .DestroyPipelineLayout = vn_DestroyPipelineLayout,
    .CreateSampler = vn_CreateSampler,
    .DestroySampler = vn_DestroySampler,
    .CreateDescriptorSetLayout = vn_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = vn_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = vn_CreateDescriptorPool,
    .DestroyDescriptorPool = vn_DestroyDescriptorPool,
    .ResetDescriptorPool = vn_ResetDescriptorPool,
    .AllocateDescriptorSets = vn_AllocateDescriptorSets,
    .FreeDescriptorSets = vn_FreeDescriptorSets,
    .UpdateDescriptorSets = vn_UpdateDescriptorSets,
    .CreateFramebuffer = vn_CreateFramebuffer,
    .DestroyFramebuffer = vn_DestroyFramebuffer,
    .CreateRenderPass = vn_CreateRenderPass,
    .DestroyRenderPass = vn_DestroyRenderPass,
    .GetRenderAreaGranularity = vn_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = vn_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = vn_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = vn_CreateCommandPool,
    .DestroyCommandPool = vn_DestroyCommandPool,
    .ResetCommandPool = vn_ResetCommandPool,
    .AllocateCommandBuffers = vn_AllocateCommandBuffers,
    .FreeCommandBuffers = vn_FreeCommandBuffers,
    .BeginCommandBuffer = vn_BeginCommandBuffer,
    .EndCommandBuffer = vn_EndCommandBuffer,
    .ResetCommandBuffer = vn_ResetCommandBuffer,
    .CmdBindPipeline = vn_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = vn_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = vn_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = vn_CmdSetViewport,
    .CmdSetScissor = vn_CmdSetScissor,
    .CmdSetLineWidth = vn_CmdSetLineWidth,
    .CmdSetDepthBias = vn_CmdSetDepthBias,
    .CmdSetBlendConstants = vn_CmdSetBlendConstants,
    .CmdSetDepthBounds = vn_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = vn_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = vn_CmdSetStencilWriteMask,
    .CmdSetStencilReference = vn_CmdSetStencilReference,
    .CmdBindDescriptorSets = vn_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = vn_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = vn_CmdBindVertexBuffers,
    .CmdDraw = vn_CmdDraw,
    .CmdDrawIndexed = vn_CmdDrawIndexed,
    .CmdDrawMultiEXT = vn_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = vn_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = vn_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = vn_CmdDrawIndexedIndirect,
    .CmdDispatch = vn_CmdDispatch,
    .CmdDispatchIndirect = vn_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = vn_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = vn_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = vn_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = vn_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = vn_CmdCopyBuffer,
    .CmdCopyImage = vn_CmdCopyImage,
    .CmdBlitImage = vn_CmdBlitImage,
    .CmdCopyBufferToImage = vn_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = vn_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = vn_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = vn_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = vn_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = vn_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = vn_CmdUpdateBuffer,
    .CmdFillBuffer = vn_CmdFillBuffer,
    .CmdClearColorImage = vn_CmdClearColorImage,
    .CmdClearDepthStencilImage = vn_CmdClearDepthStencilImage,
    .CmdClearAttachments = vn_CmdClearAttachments,
    .CmdResolveImage = vn_CmdResolveImage,
    .CmdSetEvent = vn_CmdSetEvent,
    .CmdResetEvent = vn_CmdResetEvent,
    .CmdWaitEvents = vn_CmdWaitEvents,
    .CmdPipelineBarrier = vn_CmdPipelineBarrier,
    .CmdBeginQuery = vn_CmdBeginQuery,
    .CmdEndQuery = vn_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = vn_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = vn_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = vn_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = vn_CmdResetQueryPool,
    .CmdWriteTimestamp = vn_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = vn_CmdCopyQueryPoolResults,
    .CmdPushConstants = vn_CmdPushConstants,
    .CmdBeginRenderPass = vn_CmdBeginRenderPass,
    .CmdNextSubpass = vn_CmdNextSubpass,
    .CmdEndRenderPass = vn_CmdEndRenderPass,
    .CmdExecuteCommands = vn_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = vn_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = vn_CreateSwapchainKHR,
    .DestroySwapchainKHR = vn_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = vn_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = vn_AcquireNextImageKHR,
    .QueuePresentKHR = vn_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = vn_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = vn_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = vn_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = vn_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = vn_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = vn_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = vn_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = vn_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = vn_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = vn_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = vn_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = vn_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = vn_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = vn_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = vn_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = vn_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = vn_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = vn_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = vn_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = vn_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = vn_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = vn_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = vn_CmdPushDescriptorSetKHR,
    .TrimCommandPool = vn_TrimCommandPool,
    .TrimCommandPoolKHR = vn_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = vn_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = vn_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = vn_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = vn_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = vn_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = vn_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = vn_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = vn_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = vn_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = vn_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = vn_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = vn_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = vn_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = vn_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = vn_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = vn_GetFenceFdKHR,
    .ImportFenceFdKHR = vn_ImportFenceFdKHR,
    .DisplayPowerControlEXT = vn_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = vn_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = vn_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = vn_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = vn_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = vn_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = vn_BindBufferMemory2,
    .BindBufferMemory2KHR = vn_BindBufferMemory2KHR,
    .BindImageMemory2 = vn_BindImageMemory2,
    .BindImageMemory2KHR = vn_BindImageMemory2KHR,
    .CmdSetDeviceMask = vn_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = vn_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = vn_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = vn_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = vn_AcquireNextImage2KHR,
    .CmdDispatchBase = vn_CmdDispatchBase,
    .CmdDispatchBaseKHR = vn_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = vn_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = vn_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = vn_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = vn_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = vn_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = vn_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = vn_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = vn_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = vn_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = vn_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = vn_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = vn_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = vn_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = vn_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = vn_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = vn_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = vn_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = vn_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = vn_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = vn_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = vn_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = vn_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = vn_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = vn_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = vn_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = vn_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = vn_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = vn_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = vn_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = vn_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = vn_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = vn_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = vn_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = vn_GetDeviceQueue2,
    .CreateValidationCacheEXT = vn_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = vn_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = vn_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = vn_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = vn_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = vn_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = vn_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = vn_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = vn_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = vn_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = vn_GetShaderInfoAMD,
    .SetLocalDimmingAMD = vn_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = vn_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = vn_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = vn_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = vn_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = vn_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = vn_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = vn_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = vn_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = vn_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = vn_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = vn_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = vn_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = vn_CreateRenderPass2,
    .CreateRenderPass2KHR = vn_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = vn_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = vn_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = vn_CmdNextSubpass2,
    .CmdNextSubpass2KHR = vn_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = vn_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = vn_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = vn_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = vn_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = vn_WaitSemaphores,
    .WaitSemaphoresKHR = vn_WaitSemaphoresKHR,
    .SignalSemaphore = vn_SignalSemaphore,
    .SignalSemaphoreKHR = vn_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = vn_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = vn_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = vn_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = vn_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = vn_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = vn_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = vn_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = vn_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = vn_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = vn_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = vn_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = vn_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = vn_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = vn_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = vn_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = vn_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = vn_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = vn_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = vn_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = vn_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = vn_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = vn_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = vn_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = vn_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = vn_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = vn_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = vn_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = vn_CompileDeferredNV,
    .CreateAccelerationStructureNV = vn_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = vn_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = vn_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = vn_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = vn_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = vn_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = vn_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = vn_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = vn_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = vn_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = vn_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = vn_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = vn_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = vn_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = vn_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = vn_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = vn_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = vn_CmdTraceRaysKHR,
    .CmdTraceRaysNV = vn_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = vn_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = vn_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = vn_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = vn_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = vn_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = vn_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = vn_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = vn_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = vn_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = vn_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = vn_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = vn_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = vn_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = vn_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = vn_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = vn_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = vn_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = vn_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = vn_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = vn_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = vn_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = vn_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = vn_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = vn_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = vn_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = vn_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = vn_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = vn_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = vn_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = vn_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = vn_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = vn_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = vn_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = vn_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = vn_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = vn_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = vn_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = vn_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = vn_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = vn_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = vn_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = vn_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = vn_CmdSetLineStipple,
    .CmdSetLineStippleKHR = vn_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = vn_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = vn_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = vn_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = vn_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = vn_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = vn_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = vn_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = vn_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = vn_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = vn_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = vn_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = vn_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = vn_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = vn_AntiLagUpdateAMD,
    .CmdSetCullMode = vn_CmdSetCullMode,
    .CmdSetCullModeEXT = vn_CmdSetCullModeEXT,
    .CmdSetFrontFace = vn_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = vn_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = vn_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = vn_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = vn_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = vn_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = vn_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = vn_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = vn_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = vn_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = vn_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = vn_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = vn_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = vn_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = vn_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = vn_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = vn_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = vn_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = vn_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = vn_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = vn_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = vn_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = vn_CmdSetStencilOp,
    .CmdSetStencilOpEXT = vn_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = vn_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = vn_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = vn_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = vn_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = vn_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = vn_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = vn_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = vn_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = vn_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = vn_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = vn_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = vn_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = vn_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = vn_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = vn_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = vn_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = vn_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = vn_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = vn_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = vn_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = vn_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = vn_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = vn_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = vn_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = vn_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = vn_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = vn_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = vn_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = vn_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = vn_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = vn_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = vn_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = vn_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = vn_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = vn_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = vn_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = vn_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = vn_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = vn_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = vn_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = vn_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = vn_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = vn_DestroyPrivateDataSlotEXT,
    .SetPrivateData = vn_SetPrivateData,
    .SetPrivateDataEXT = vn_SetPrivateDataEXT,
    .GetPrivateData = vn_GetPrivateData,
    .GetPrivateDataEXT = vn_GetPrivateDataEXT,
    .CmdCopyBuffer2 = vn_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = vn_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = vn_CmdCopyImage2,
    .CmdCopyImage2KHR = vn_CmdCopyImage2KHR,
    .CmdBlitImage2 = vn_CmdBlitImage2,
    .CmdBlitImage2KHR = vn_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = vn_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = vn_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = vn_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = vn_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = vn_CmdResolveImage2,
    .CmdResolveImage2KHR = vn_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = vn_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = vn_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = vn_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = vn_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = vn_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = vn_CmdSetEvent2,
    .CmdSetEvent2KHR = vn_CmdSetEvent2KHR,
    .CmdResetEvent2 = vn_CmdResetEvent2,
    .CmdResetEvent2KHR = vn_CmdResetEvent2KHR,
    .CmdWaitEvents2 = vn_CmdWaitEvents2,
    .CmdWaitEvents2KHR = vn_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = vn_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = vn_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = vn_QueueSubmit2,
    .QueueSubmit2KHR = vn_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = vn_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = vn_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = vn_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = vn_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = vn_CopyMemoryToImage,
    .CopyMemoryToImageEXT = vn_CopyMemoryToImageEXT,
    .CopyImageToMemory = vn_CopyImageToMemory,
    .CopyImageToMemoryEXT = vn_CopyImageToMemoryEXT,
    .CopyImageToImage = vn_CopyImageToImage,
    .CopyImageToImageEXT = vn_CopyImageToImageEXT,
    .TransitionImageLayout = vn_TransitionImageLayout,
    .TransitionImageLayoutEXT = vn_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = vn_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = vn_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = vn_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = vn_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = vn_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = vn_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = vn_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = vn_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = vn_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = vn_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = vn_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = vn_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = vn_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = vn_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = vn_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = vn_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = vn_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = vn_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = vn_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = vn_CreateCuModuleNVX,
    .CreateCuFunctionNVX = vn_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = vn_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = vn_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = vn_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = vn_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = vn_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = vn_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = vn_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = vn_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = vn_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = vn_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = vn_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = vn_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = vn_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = vn_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = vn_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = vn_WaitForPresent2KHR,
    .WaitForPresentKHR = vn_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = vn_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = vn_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = vn_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = vn_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = vn_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = vn_CmdBeginRendering,
    .CmdBeginRenderingKHR = vn_CmdBeginRenderingKHR,
    .CmdEndRendering = vn_CmdEndRendering,
    .CmdEndRendering2KHR = vn_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = vn_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = vn_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = vn_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = vn_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = vn_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = vn_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = vn_BuildMicromapsEXT,
    .DestroyMicromapEXT = vn_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = vn_CmdCopyMicromapEXT,
    .CopyMicromapEXT = vn_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = vn_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = vn_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = vn_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = vn_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = vn_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = vn_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = vn_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = vn_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = vn_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = vn_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = vn_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = vn_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = vn_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = vn_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = vn_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = vn_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = vn_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = vn_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = vn_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = vn_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = vn_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = vn_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = vn_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = vn_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = vn_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = vn_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = vn_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = vn_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = vn_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = vn_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = vn_MapMemory2,
    .MapMemory2KHR = vn_MapMemory2KHR,
    .UnmapMemory2 = vn_UnmapMemory2,
    .UnmapMemory2KHR = vn_UnmapMemory2KHR,
    .CreateShadersEXT = vn_CreateShadersEXT,
    .DestroyShaderEXT = vn_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = vn_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = vn_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = vn_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = vn_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = vn_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = vn_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = vn_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = vn_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = vn_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = vn_CmdPushConstants2,
    .CmdPushConstants2KHR = vn_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = vn_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = vn_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = vn_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = vn_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = vn_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = vn_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = vn_SetLatencySleepModeNV,
    .LatencySleepNV = vn_LatencySleepNV,
    .SetLatencyMarkerNV = vn_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = vn_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = vn_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = vn_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = vn_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = vn_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = vn_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = vn_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = vn_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = vn_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = vn_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = vn_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = vn_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = vn_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = vn_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = vn_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = vn_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = vn_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = vn_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = vn_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = vn_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = vn_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = vn_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = vn_CreateTensorARM,
    .DestroyTensorARM = vn_DestroyTensorARM,
    .CreateTensorViewARM = vn_CreateTensorViewARM,
    .DestroyTensorViewARM = vn_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = vn_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = vn_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = vn_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = vn_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = vn_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = vn_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = vn_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = vn_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = vn_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = vn_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = vn_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = vn_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = vn_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = vn_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = vn_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = vn_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = vn_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = vn_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = vn_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = vn_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = vn_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = vn_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = vn_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = vn_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = vn_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = vn_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = vn_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = vn_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = vn_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = vn_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = vn_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = vn_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = vn_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = vn_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = vn_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = vn_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = vn_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = vn_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = vn_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = vn_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = vn_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = vn_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = vn_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = vn_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = vn_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = vn_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = vn_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = vn_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = vn_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = vn_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = vn_CmdSetDispatchParametersARM,
};


